# -*- coding: utf-8 -*-
"""
Created on Fri Feb 12 16:52:22 2021

@author: chris.kerklaan
"""

# First-party imports
import logging

# Third-party imports
import shapely
from osgeo import ogr, _ogr

# Local imports
from .fields import Fields, GeometryFields
from . import Point, MultiPoint, LineString, MultiLineString, Polygon, MultiPolygon

# GLOBALS
logger = logging.getLogger(__name__)
GEOMETRIES = [
    Point,
    MultiPoint,
    LineString,
    MultiLineString,
    Polygon,
    MultiPolygon,
]
OGR_GEOMETRY_CONVERSION = {}
for geometry in GEOMETRIES:
    OGR_GEOMETRY_CONVERSION.update(
        {coverage: geometry for coverage in geometry.ogr_coverage}
    )


class Feature:
    """wrapper of ogr feature

    Edits to geometry and fields are never set, just do object['field']
    to set the field
    """

    def __init__(self, feature: ogr.Feature, layer: ogr.Layer, fields: Fields):
        self.feature = feature

        if feature is not None:
            self.exists = True
        else:
            layer.ResetReading()
            first = None
            if layer.GetFeatureCount() > 0:
                first = layer.GetNextFeature().GetFID()
            raise IndexError(
                f"Feature does not exist, starts at: {first}, use vector.fids for all."
            )

        self.layer = layer
        self.fields = fields

    def __iter__(self):
        for key, value in self.items.items():
            yield key, value

    def __setitem__(self, field_name, value):
        _ogr.Feature_SetField(self.feature, field_name, value)
        _ogr.Layer_SetFeature(self.layer, self.feature)

    def __getitem__(self, field_name):
        return self.get_item(field_name)

    def __str__(self):
        return str(self.all_items)

    def __repr__(self):
        return str(self.all_items)

    def copy(self):
        return Feature(_ogr.Feature_Clone(self.feature))

    @property
    def all_items(self):
        return {"items": self.items, "geometry": self.geometry}

    @property
    def fid(self):
        return _ogr.Feature_GetFID(self.feature)

    @fid.setter
    def fid(self, value):
        _ogr.Feature_SetFID(value)
        _ogr.Layer_SetFeature(self.layer, self.feature)

    def get_item(self, field_name):
        i = self.fields.indices[field_name]
        if _ogr.Feature_IsFieldNull(self.feature, i):
            return None
        else:
            func = self.fields.get_field_func(field_name)
            return func(self.feature, i)

    def get_geometry(self):
        geometry = _ogr.Feature_GetGeometryRef(self.feature)
        if not geometry:
            geometry = _ogr.Feature_GetGeomFieldRef(self.feature, 1)
        return geometry

    @property
    def geometry(self):
        return geometry_factory(self.get_geometry())

    @geometry.setter
    def geometry(self, value):
        if isinstance(value, shapely.geometry.base.BaseGeometry):
            value = ogr.CreateGeometryFromWkb(value.wkb)

        _ogr.Feature_SetGeometry(self.feature, value)
        _ogr.Layer_SetFeature(self.layer, self.feature)

    @property
    def items(self):
        items = {}
        for field_name in self.fields.names:
            items[field_name] = self[field_name]
        return items

    @property
    def id(self):
        return self.fid

    @property
    def spatial_reference(self):
        return self.geometry.spatial_reference

    @property
    def area(self):
        return self.geometry.area

    @property
    def geometry_fields(self):
        return GeometryFields(self.feature)

    def simplify(self, wkb_threshold, start=10000000, quiet=False):
        """simplifying a geometry to a wkb size threshold"""
        geometry = self.geometry
        while geometry.WkbSize() > wkb_threshold:
            geometry = geometry.simply(1 / start)
            start = start / 10
        return geometry

    def reproject(self, out_epsg):
        return self.geometry.reproject(out_epsg)


def to_ogr(feature):
    """Returns an ogr feature"""
    if hasattr(feature, "feature"):
        return feature.feature
    return feature


def geometry_type_class(ogr_geometry_type):
    """returns the correct geometry type given an ogr_geometry"""
    try:
        geometry_class = OGR_GEOMETRY_CONVERSION[ogr_geometry_type]
    except KeyError:
        name = ogr.GeometryTypeToName(ogr_geometry_type)
        raise ValueError(f"Found unexpected geometry type {name}")
    else:
        return geometry_class


def geometry_factory(geometry: ogr.Geometry):
    """converts an ogr geometry to the custom ogr geometry classes"""
    if geometry:
        geometry_type = _ogr.Geometry_GetGeometryType(geometry)
        return geometry_type_class(geometry_type)(geometry)



# -*- coding: utf-8 -*-
"""
Created on Tue Apr  6 14:01:54 2021

@author: chris.kerklaan

# settings: https://github.com/nens/threedi-turtle-scripts/blob/master/sql-scripts/scripts-2B/process-3di/settings_for_2d_model.sql
"""
# system imports
import copy

# Local imports
from .constants import Properties

PROPERTIES = Properties()


template_integrated = {
    "global_settings": {
        # General
        "id": None,
        "name": "test",
        "use_2d_flow": 1,
        "use_1d_flow": 1,
        "use_0d_inflow": 1,
        "use_2d_rain": 1,
        # Grid
        "grid_space": 10,
        "kmax": 1,
        "table_step_size": 0.01,
        # Terrain info
        # Dem
        "dem_file": None,
        "epsg_code": 28992,
        # Friction
        "frict_coef_file": None,
        "frict_coef": 0.026,
        "frict_type": 2,
        "frict_avg": 0,
        # Groundwater
        "initial_groundwater_level_file": None,
        "initial_groundwater_level": None,
        "initial_groundwater_level_type": None,
        # Initial waterlevel
        "initial_waterlevel_file": None,
        "initial_waterlevel": -10,
        "water_level_ini_type": None,
        # Interception
        "interception_file": None,
        "interception_global": None,
        # Wind
        "wind_shielding_file": None,
        # Time
        "start_date": "2014-05-14",
        "start_time": "2014-05-14 17:42:18",
        "sim_time_step": 5,
        "timestep_plus": True,
        "minimum_sim_time_step": 1,
        "maximum_sim_time_step": 60,
        "nr_timesteps": 2880,
        "output_time_step": 1,
        # Settings id's
        "interflow_settings_id": None,
        "groundwater_settings_id": None,
        "numerical_settings_id": 1,
        "simple_infiltration_settings_id": 1,
        "control_group_id": None,
        # Extra options 1D
        "advection_1d": 0,
        "dist_calc_points": 1000,
        "manhole_storage_area": 100,
        "max_angle_1d_advection": 90,
        "table_step_size_1d": None,
        "advection_2d": 0,
        "dem_obstacle_detection": False,
        "guess_dams": False,
        "dem_obstacle_height": None,
        "embedded_cutoff_threshold": None,
        "flooding_threshold": 0.001,
        "table_step_size_volume_2d": None,
    },
    "numerical_settings": {
        # General
        "id": None,
        "limiter_grad_1d": None,
        "limiter_grad_2d": None,
        "limiter_slope_crossectional_area_2d": None,
        "limiter_slope_friction_2d": None,
        # Matrix
        "convergence_cg": None,
        "convergence_eps": 0.00005,
        "use_of_cg": 20,
        "use_of_nested_newton": 1,
        "max_degree": 7,
        "max_nonlin_iterations": 20,
        "precon_cg": 1,
        "integration_method": 0,
        # Thresholds
        "flow_direction_threshold": None,
        "general_numerical_threshold": None,
        "thin_water_layer_definition": None,
        "minimum_friction_velocity": None,
        "minimum_surface_area": None,
        # Miscellaneous
        "cfl_strictness_factor_1d": 1,
        "cfl_strictness_factor_2d": 1,
        "frict_shallow_water_correction": 0,
        "pump_implicit_ratio": 1,
        "preissmann_slot": None,
    },
    "simple_infiltration": {
        "id": None,
        "display_name": "Infiltration",
        "infiltration_rate": 0.01,
        "infiltration_rate_file": None,
        "max_infiltration_capacity_file": None,
        "infiltration_surface_option": 0,
    },
    "connection_nodes": {
        "id": None,
        "code": None,
        "initial_waterlevel": None,
        "storage_area": 0.64,
        "the_geom": None,
    },
    "manhole": {
        # general
        "id": None,
        "display_name": None,
        "code": None,
        "connection_node_id": None,
        "calculation_type": None,
        "bottom_level": None,
        "calculation_type": None,
        "drain_level": None,
        "length": None,
        "width": None,
        "zoom_category": 1,
        "manhole_indicator": None,
        "shape": None,
        "surface_level": None,
        "sediment_level": None,
    },
    "channel": {
        "id": None,
        "calculation_type": 102,
        "code": None,
        "connection_node_end_id": None,
        "connection_node_start_id": None,
        "display_name": None,
        "dist_calc_points": 10,
        "zoom_category": 5,
        "the_geom": None,
    },
    "cross_section_definition": {
        "id": None,
        "code": None,
        "height": None,
        "shape": None,
        "width": None,
    },
    "cross_section_location": {
        "id": None,
        "bank_level": None,
        "channel_id": None,
        "code": None,
        "definition_id": None,
        "friction_type": 2,
        "friction_value": 0.026,
        "reference_level": None,
        "the_geom": None,
    },
    "grid_refinement_area": {
        "id": None,
        "display_name": None,
        "refinement_level": None,
        "the_geom": None,
        "code": None,
    },
    "grid_refinement": {
        "id": None,
        "display_name": None,
        "refinement_level": None,
        "the_geom": None,
        "code": None,
    },
    "1d_boundary_conditions": {
        "id": None,
        "boundary_type": None,
        "connection_node_id": None,
        "timeseries": None,
    },
    "pumpstation": {
        "id": None,
        "capacity": None,
        "zoom_category": 5,
        "code": None,
        "connection_node_end_id": None,
        "connection_node_start_id": None,
        "classification": None,
        "display_name": None,
        "lower_stop_level": None,
        "sewerage": False,
        "start_level": None,
        "type": 1,
        "upper_stop_level": None,
    },
    "levee": {
        "id": None,
        "crest_level": None,
        "material": 2,
        "max_breach_depth": 10,
        "code": None,
        "the_geom": None,
    },
    "obstacle": {
        "id": None,
        "crest_level": None,
        "code": None,
        "the_geom": None,
    },
    "interflow": {
        "id": None,
        "display_name": None,
        "interflow_type": None,
        "porosity": None,
        "porosity_file": 10,
        "porosity_layer_thickness": None,
        "hydraulic_conductivity_file": None,
        "hydraulic_conductivity": None,
        "impervious_layer_elevation": None,
    },
    "orifice": {
        "display_name": None,
        "code": None,
        "zoom_category": None,
        "connection_node_start_id": None,
        "connection_node_end_id": None,
        "crest_level": None,
        "crest_type": 4,
        "cross_section_definition_id": None,
        "discharge_coefficient_negative": 0.8,
        "discharge_coefficient_positive": 0.8,
        "friction_type": 2,
        "friction_value": 0.011,
        "sewerage": 1,
    },
}

template_2d_model = {
    "global_settings": {
        # General
        "id": None,
        "name": "test",
        "use_2d_flow": 1,
        "use_1d_flow": 0,
        "use_0d_inflow": 0,
        "use_2d_rain": 1,
        # Grid
        "grid_space": 10,
        "kmax": 1,
        "table_step_size": 0.01,
        # Terrain info
        # Dem
        "dem_file": "rasters/dem.tif",
        "epsg_code": 28992,
        # Friction
        "frict_coef_file": None,
        "frict_coef": 0.026,
        "frict_type": 2,
        "frict_avg": 0,
        # Groundwater
        "initial_groundwater_level_file": None,
        "initial_groundwater_level": None,
        "initial_groundwater_level_type": None,
        # Initial waterlevel
        "initial_waterlevel_file": None,
        "initial_waterlevel": -10,
        "water_level_ini_type": None,
        # Interception
        "interception_file": None,
        "interception_global": None,
        # Wind
        "wind_shielding_file": None,
        # Time
        "start_date": "2014-05-14",
        "start_time": "2014-05-14 17:42:18",
        "sim_time_step": 30,
        "timestep_plus": True,
        "minimum_sim_time_step": 30,
        "maximum_sim_time_step": 300,
        "nr_timesteps": 240,
        "output_time_step": 300,
        # Settings id's
        "interflow_settings_id": None,
        "groundwater_settings_id": None,
        "numerical_settings_id": 1,
        "simple_infiltration_settings_id": 1,
        "control_group_id": None,
        # Extra options 1D
        "advection_1d": 0,
        "dist_calc_points": 1000,
        "manhole_storage_area": 0,
        "max_angle_1d_advection": 90,
        "table_step_size_1d": None,
        "advection_2d": 0,
        "dem_obstacle_detection": False,
        "guess_dams": False,
        "dem_obstacle_height": None,
        "embedded_cutoff_threshold": None,
        "flooding_threshold": 0.01,
        "table_step_size_volume_2d": None,
    },
    "numerical_settings": {
        "id": None,
        # General
        "limiter_grad_1d": None,
        "limiter_grad_2d": None,
        "limiter_slope_crossectional_area_2d": None,
        "limiter_slope_friction_2d": None,
        # Matrix
        "convergence_cg": None,
        "convergence_eps": 0.00001,
        "use_of_cg": 20,
        "use_of_nested_newton": 0,
        "max_degree": 5,
        "max_nonlin_iterations": 20,
        "precon_cg": 1,
        "integration_method": 0,
        # Thresholds
        "flow_direction_threshold": None,
        "general_numerical_threshold": None,
        "thin_water_layer_definition": None,
        "minimum_friction_velocity": None,
        "minimum_surface_area": None,
        # Miscellaneous
        "cfl_strictness_factor_1d": 1,
        "cfl_strictness_factor_2d": 1,
        "frict_shallow_water_correction": 0,
        "pump_implicit_ratio": 1,
        "preissmann_slot": None,
    },
    "simple_infiltration": {
        "id": None,
        "display_name": "Infiltration",
        "infiltration_rate": 0,
        "infiltration_rate_file": None,
        "max_infiltration_capacity_file": None,
        "infiltration_surface_option": 0,
    },
    "connection_nodes": {
        "id": None,
        "code": None,
        "initial_waterlevel": None,
        "storage_area": 0.64,
        "the_geom": None,
    },
    "manhole": {
        "id": None,
        # general
        "display_name": None,
        "code": None,
        "connection_node_id": None,
        "calculation_type": None,
    },
    "grid_refinement": {
        "id": None,
        "display_name": None,
        "refinement_level": None,
        "the_geom": None,
        "code": None,
    },
    "grid_refinement_area": {
        "id": None,
        "display_name": None,
        "refinement_level": None,
        "the_geom": None,
    },
    "1d_boundary_conditions": {
        "id": None,
        "boundary_type": None,
        "connection_node_id": None,
        "timeseries": None,
    },
    "levee": {
        "id": None,
        "crest_level": None,
        "material": None,
        "code": None,
        "max_breach_depth": 10,
        "the_geom": None,
    },
    "obstacle": {
        "id": None,
        "crest_level": None,
        "code": None,
        "the_geom": None,
    },
    "interflow": {
        "id": None,
        "display_name": None,
        "interflow_type": None,
        "porosity": None,
        "porosity_file": 10,
        "porosity_layer_thickness": None,
        "hydraulic_conductivity_file": None,
        "hydraulic_conductivity": None,
        "impervious_layer_elevation": None,
    },
    "orifice": {
        "display_name": None,
        "code": None,
        "zoom_category": None,
        "connection_node_start_id": None,
        "connection_node_end_id": None,
        "crest_level": None,
        "crest_type": 4,
        "cross_section_definition_id": None,
        "discharge_coefficient_negative": 0.8,
        "discharge_coefficient_positive": 0.8,
        "friction_type": 2,
        "friction_value": 0.011,
        "sewerage": 1,
    },
}


template_mapping = {
    "integrated": copy.deepcopy(template_integrated),
    "2d_model": copy.deepcopy(template_2d_model),
}


class ThreediString(str):
    """
    Class acts (therefore we use __new__) as a string, with multiple inputs
    Class has info (__init__)
    Class prints info

    """

    name = "ThreediString"

    def __new__(cls, content, info, table):
        return super().__new__(cls, content)

    def __init__(self, content, info, table):
        self.table = table
        self.finfo = info
        self.content = content

    @property
    def info(self):
        print(self.finfo)

    def __str__(self):
        if self.info is not None:
            return f"Value: {self.content}\nDescription: {self.info}"
        elif self.content is not None:
            return str(self.content)
        else:
            return "(ThreediString) Template not filled"

    def __repr__(self):
        if self.info is not None:
            return f"Value: {self.content}\nDescription: {self.info}"
        elif self.content is not None:
            return str(self.content)
        else:
            return "(ThreediString) Template not filled"


class ThreediInteger(int):
    """integer or bool"""

    name = "ThreediInteger"

    def __new__(cls, content, info, table):
        return super().__new__(cls, content)

    def __init__(self, content, info, table):
        self.table = table
        self.finfo = info
        self.content = content

    @property
    def info(self):
        print(self.finfo)

    def __str__(self):
        if self.info is not None:
            return f"Value: {self.content}\nDescription: {self.info}"
        elif self.content is not None:
            return str(self.content)
        else:
            return "(ThreediInteger) Template not filled"

    def __repr__(self):
        if self.info is not None:
            return f"Value: {self.content}\nDescription: {self.info}"
        elif self.content is not None:
            return str(self.content)
        else:
            return "(ThreediInteger) Template not filled"


class ThreediFloat(float):
    """Float"""

    name = "ThreediFloat"

    def __new__(cls, content, info, table):
        return super().__new__(cls, content)

    def __init__(self, content, info, table):
        self.table = table
        self.finfo = info
        self.content = content

    @property
    def info(self):
        print(self.finfo)

    def __str__(self):
        if self.info is not None:
            return f"Value: {self.content}\nDescription: {self.info}"
        elif self.content is not None:
            return str(self.content)
        else:
            return "(ThreediFloat) Template not filled"

    def __repr__(self):
        if self.info is not None:
            return f"Value: {self.content}\nDescription: {self.info}"
        elif self.content is not None:
            return str(self.content)
        else:
            return "(ThreediFloat) Template not filled"


def threedi_value(value, info, table):
    value_type = type(value)
    if value_type == str or value_type == type(None):
        return ThreediString(value, info, table)
    elif value_type == int or value_type == bool:
        return ThreediInteger(value, info, table)
    elif value_type == float:
        return ThreediFloat(value, info, table)
    else:
        return value


class ThreediTemplateTable(dict):
    """table template"""

    def __init__(self, in_dict):
        self.__dict__ = in_dict

    def __setitem__(self, key, value):
        d = dict(self.__dict__)
        table = self.table
        # if type(d[key]) != type(value):
        #     raise TypeError(f"Type mismatch {type(d[key]) } {type(value)}")

        d[key] = threedi_value(
            value, PROPERTIES["v2_" + table][key]["description"], table
        )
        self.__dict__ = d

    def __getitem__(self, key):
        return self.__dict__[key]

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return str(self.__dict__)

    def items(self):
        for key, value in self.__dict__.items():
            if not key == "table":
                yield key, value

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def get(self, key):
        try:
            value = self[key]
        except KeyError:
            value = None
        return value


class Templates(object):
    def __init__(self, template_type="integrated"):
        self.type = template_type
        self.template = template_mapping[self.type]

    def get_template(self, table):
        values = self.template[table]
        new_dict = {}
        for key, value in values.items():
            v = threedi_value(
                value, PROPERTIES["v2_" + table][key]["description"], table
            )
            new_dict[key] = v

        table_template = ThreediTemplateTable(new_dict)
        table_template.table = table
        return table_template

    @property
    def global_setting(self):
        return self.get_template("global_settings")

    @property
    def numerical_setting(self):
        return self.get_template("numerical_settings")

    @property
    def simple_infiltration(self):
        return self.get_template("simple_infiltration")

    @property
    def node(self):
        return self.get_template("connection_nodes")

    @property
    def manhole(self):
        return self.get_template("manhole")

    @property
    def channel(self):
        return self.get_template("channel")

    @property
    def cross_section(self):
        return self.get_template("cross_section_definition")

    @property
    def cross_section_location(self):
        return self.get_template("cross_section_location")

    @property
    def grid_refinement_area(self):
        return self.get_template("grid_refinement_area")

    @property
    def grid_refinement(self):
        return self.get_template("grid_refinement")

    @property
    def boundary_condition_1d(self):
        return self.get_template("1d_boundary_conditions")

    @property
    def pumpstation(self):
        return self.get_template("pumpstation")

    @property
    def levee(self):
        return self.get_template("levee")

    @property
    def interflow(self):
        return self.get_template("interflow")

    @property
    def obstacle(self):
        return self.get_template("obstacle")

    @property
    def orifice(self):
        return self.get_template("orifice")



# -*- coding: utf-8 -*-
"""
Created on Mon Nov  2 08:14:27 2020

@author: chris.kerklaan

#TODO:
    1. Set data constraints to look like the empty template
    1. Checks
        1. Data-checks is something allow or not-allowed in the field,
            These checks must be done to run smoothly through inpy            
        2. 
    General:
        Add functions
    3. Create extension of ogr feature here, to acces set_items and perform checks there
    4. Specify input and output of different variables.
    5. Connect get items field to fields of constants
    
Implemented checks:
    1. value type check in input feature at __setitem__ of feature, and custom in model
    
"""

# system imports
import operator
import logging

# Third-party imports
import json
import datetime as dt

# local imports
from threedi_edits import LineString
from threedi_edits.threedi.vector import ThreediVector
from threedi_edits.threedi.constants.templates import Templates
from threedi_edits.utils.project import Classes, Functions
from threedi_edits.threedi.utils.conversions import METER_TO_DEGREE

# globals
TEMPLATE = Templates()

# logger
logger = logging.getLogger(__name__)

# structure
classes = Classes(__name__, local_only=True)
functions = Functions(__name__)


class SimpleInfiltration(ThreediVector):
    constraints = {}

    def __init__(self, ds, layer_name):
        ThreediVector.__init__(self, ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class GlobalSettings(ThreediVector):
    def __init__(self, ds, layer_name):
        ThreediVector.__init__(self, ds, layer_name)

    def check_scenario_exists(self):
        try:
            self[1]
        except IndexError:
            raise IndexError(
                f"""Scenario {self.scenario} does not exist, use self.set_scenario if a different one is needed"""
            )

    def custom_set_function(self, key, value):
        if key == "start_time":
            """start time must be in iso format"""
            if value is not None:
                try:
                    value = value.replace("/", "-")
                    dt_object = dt.datetime.fromisoformat(value)
                    value = dt_object.isoformat()
                except Exception:
                    raise ValueError(
                        "Object did not successfully convert to isoformat for start_time",
                        "is the input something like: 2014-05-14 17:42:18?",
                    )
        if key == "grid_space":
            """gridspace must be at least twice the cellsize of a raster"""
            if self.dem is not None:
                try:
                    resolution = self.dem.resolution["width"]
                    grid_space = float(self[self.scenario]["grid_space"])
                    if not ((grid_space / resolution) >= 2):
                        raise ValueError(
                            f"""
                                         Grid space must be at least twice 
                                         the resolution of the dem
                                         current grid_space {grid_space},
                                         current resolution of dem {resolution} 
                                         """
                        )
                except Exception:
                    pass
        return key, value


class AggregationSettings(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class Interflow(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class Groundwater(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class BoundaryConditions1d(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        """also accepts a list of tuplees"""

        if key == "timeseries":

            if type(value) == list:
                value = "\n".join([f"{_min},{ts}" for _min, ts in value])

        return key, value

    def to_dict(self, interpolate=False, path=None):
        json_bcs = []
        for bc in self:
            json_bc = {"interpolate": interpolate, "values": []}
            for values in bc["timeseries"].split("\n"):
                json_bc["values"] = [float(i) for i in values.split(",")]
            json_bcs.append((bc["connection_node_id"], json_bc))

        json_bcs_sorted = sorted(json_bcs, key=lambda x: x[0])
        json_bcs = [json_bc[1] for json_bc in json_bcs_sorted]

        if path:
            json_string = json.dumps(json_bcs)
            json_file = open(path, "w")
            json_file.write(json_string)
            json_file.close()

        return json_bcs


class BoundaryConditions2d(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class Lateral1d(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class Lateral2d(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class CalculationPoint(ThreediVector):
    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class ConnectionNodes(ThreediVector):
    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value

    def custom_add_function(self, fid, items, geometry):
        if geometry:
            if len(self.spatial_filter(geometry)) > 0:
                raise ValueError(
                    f""" 
                      Trying to add a double geometry
                      feature 1: {fid}
                      feature 2: {self.spatial_filter(geometry)[0].fid}
                                 """
                )

        return fid, items, geometry

    def add_linestring(self, branches, template, NODES_STORAGE_AREA):
        """add connection nodes based on the start and end vertices of the channels"""

        for branch in branches:
            geometry = branch.geometry
            for geom, name in [
                (geometry.start_point, "start"),
                (geometry.end_point, "end"),
            ]:

                node = TEMPLATE.node
                node["the_geom"] = geom
                node["code"] = None
                node["storage_area"] = NODES_STORAGE_AREA
                geom.spatial_reference = branches.epsg

                try:
                    self.add(node)
                except ValueError:  # double connection
                    raise ValueError("found double connection")

    def create_line(self, node_id1: int, node_id2: int):
        """returns a geometry between two nodes"""

        return LineString.from_points(
            [self[node_id1].geometry.point, self[node_id2].geometry.point], epsg=4326
        )

    def reproject(self, epsg):
        return ConnectionNodes(self.reproject(epsg).ds, "v2_connection_nodes")


class ConnectedPnt(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class CrossSectionLocation(ThreediVector):
    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        # if self.channels != None:
        #     if key == "the_geom":
        #         filtered = self.channels.spatial_filter(value)
        #         if len(filtered) == 0:
        #             raise ValueError(
        #                 "Cross section location does not intersect a channel"
        #             )

        #         on_vertice = False
        #         for point in filtered[0].geometry.points:
        #             if point == value.point:
        #                 on_vertice = True
        #                 break
        #         if not on_vertice:
        #             raise ValueError(
        #                 "Cross section location geometry not on vertice"
        #             )

        return key, value

    def check_levels(self):
        """returns differences between the bank level and the reference level"""
        differences = []
        for location in self:
            differences.append(
                float(location["bank_level"]) - float(location["reference_level"])
            )
        return differences


class Culvert(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class Channel(ThreediVector):
    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)
        # self.csl = cross_section_locations

    def custom_set_function(self, key, value):
        return key, value

    def check_csl_presence(self):
        """checks if all channels have a cross section locations,
        returns the ids without one
        """
        channels_wo_locations = []
        for channel in self:
            if len(self.csl.filter(channel_id=channel.id)) == 0:
                channels_wo_locations.append(channel.id)
        return channels_wo_locations

    def check_start_end_point(self, minimum_distance=0.26):
        """checks if location is near the start or end point of channels"""
        # cannot be on start or end point of channel
        near_points = []
        distance = minimum_distance * METER_TO_DEGREE
        for location in self.csl:
            channel_id = int(location["channel_id"])
            channel = self[channel_id]
            near_start = (
                location.geometry.Distance(channel.geometry.start_point) < distance
            )
            near_end = location.geometry.Distance(channel.geometry.end_point) < distance

            if near_start or near_end:
                near_points.append(location.id)

        return near_points

    def check_vertices(self):
        """checks if all cross locations are on the channel vertices,
        returns the ids of locations which aer not
        """
        faulty_locations = []
        for location in self.csl:
            filtered = self.spatial_filter(location.geometry)
            if len(filtered) == 0:
                faulty_locations.append(location.id)
                logger.debug("Cross section location does not intersect a channel")
                continue

            on_vertice = False
            for point in filtered[0].geometry.points:
                if point == location.geometry.point:
                    on_vertice = True
                    break

            if not on_vertice:
                faulty_locations.append(location.id)
                logger.debug("Cross section location geometry not on vertice")

        return faulty_locations

    def closest_point_on_channel(
        self, channel_geometry, cross_geometry, minimum_distance
    ):
        """
        Checks if the geometry is located at the very start or end point of a channel,
        then it will be moved to a minimum distance
        """
        distance = minimum_distance * METER_TO_DEGREE
        near_start = cross_geometry.Distance(channel_geometry.start_point) < distance
        near_end = cross_geometry.Distance(channel_geometry.end_point) < distance
        on_vertice = cross_geometry.point in channel_geometry.points

        if near_start or near_end or not on_vertice:
            points = channel_geometry.points_on_line(
                interval=distance, start=False, end=False
            )

            closest = []
            for idx, point in enumerate(points):
                closest.append((idx, point.Distance(cross_geometry)))

            return points[sorted(closest, key=lambda x: x[1])[0][0]]
        else:
            return cross_geometry

    def non_existing_cross_section_locations(self):
        """returns a list of not existing locations"""
        channel_fids = self.fids
        return [l.id for l in self.csl if l["channel_id"] not in channel_fids]

    def predict_calculation_points(self, feature_id=None, projection=28992):
        """returns calculation points per feature"""
        if feature_id:
            feature = self[feature_id]
            geometry = feature.geometry
            interval = int(feature["dist_calc_points"])

            projected = LineString(geometry.reproject(projection))

            line_points = projected.points_on_line(interval, start=False, end=False)
            if len(line_points) == 0:
                line_points = [projected.start_point]

            calc_points = []
            for i, point in enumerate(line_points):
                fid = i + 1
                code = f"{feature_id}_{fid}_{feature['calculation_type']}"
                calc_points.append(
                    {
                        "content_type_id": feature_id,
                        "user_ref": f"{code}#{feature_id}#v2_channel#{fid}",
                        "calc_type": feature["calculation_type"] - 100,
                        "the_geom": point.reproject(4326),
                    }
                )
            return calc_points


class Pipe(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class Weir(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class Orifice(ThreediVector):
    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class Manhole(ThreediVector):
    constraints = {"drain_level": {"comparison": (operator.ge, "bottom_level")}}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value

    def set_drain_level(self, below_surface_level=0.15, none_only=True):
        """sets all drain levels with None to 0.15 m below surface level"""

        for feature in self:
            drain_level = feature["drain_level"]
            if none_only and feature["drain_level"] != None:
                continue

            feature["drain_level"] = feature["surface_level"] - below_surface_level

            self.edits.append(
                (
                    feature.fid,
                    "drain_level",
                    drain_level,
                    feature["drain_level"],
                )
            )


class Pumpstation(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class Windshielding(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class CrossSectionDefinition(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    # def custom_plot(self, feature):
    #     """this is the only way to alter this function"""

    #     def plot(filename=None):
    #         width = [float(i) for i in feature.items["width"].split(" ")]
    #         height = [float(i) for i in feature.items["height"].split(" ")]
    #         middle = max(width) / 2

    #         # step 6: make the profile ready to be plotted
    #         yz_tab_list = []
    #         for h, w in zip(height, width):
    #             if h == 0:
    #                 yz_tab_list.append([middle, h])
    #             else:
    #                 yz_tab_list.append([middle - w / 2, h])
    #                 yz_tab_list.append([middle + w / 2, h])

    #         yz_tab = np.array(yz_tab_list)
    #         yz_tab_left = yz_tab[yz_tab[:, 0] < middle]
    #         yz_tab_right = yz_tab[yz_tab[:, 0] >= middle]
    #         yz_tab_left_sorted = yz_tab_left[
    #             np.lexsort((yz_tab_left[:, 0], -yz_tab_left[:, 1]))
    #         ]
    #         yz_tab_right_sorted = yz_tab_right[
    #             np.lexsort((yz_tab_right[:, 0], yz_tab_right[:, 1]))
    #         ]
    #         yz_tab_sorted = np.concatenate(
    #             (yz_tab_left_sorted, yz_tab_right_sorted), axis=0
    #         )

    #         plt.figure(figsize=(10, 5))
    #         plt.plot(
    #             yz_tab_sorted[:, 0],
    #             yz_tab_sorted[:, 1],
    #             label="Tabulated trapezium",
    #         )
    #         plt.xlabel("Distance from bank [m]", fontdict=None, labelpad=None)
    #         plt.ylabel("Depth [m]", fontdict=None, labelpad=None)
    #         plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.0)

    #         if filename:
    #             plt.savefig(filename)
    #         plt.show()

    #     return plot

    def custom_set_function(self, key, value):
        if (key == "width" or key == "height") and type(value) == list:
            value_str = ""
            for i, v in enumerate(value):
                if i == len(value) - 1:
                    value_str = value_str + f"{v}"
                else:
                    value_str = value_str + f"{v} "
            value = value_str

        return key, value


class Obstacle(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class Levee(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class GridRefinement(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class GridRefinementArea(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        # if key == "refinement_level":
        #     if self.global_setting is not None:
        #         if value > self.global_setting['kmax']:
        #             raise ValueError("Refinment level above kmax")

        return key, value


class NumericalSettings(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class DemAverageArea(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class ImperviousSurfaceMap(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class ImperviousSurface(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class SurfaceParameters(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class SurfaceMap(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class Surface(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(key, value):
        return key, value


class Control(ThreediVector):

    constraints = {}
    required = {
        "control_group": None,
        "control_table": None,
        "control_measure_map": None,
        "control_measure_group": None,
    }

    def __init__(
        self,
        ds,
        layer_name,
        control_group=None,
        control_table=None,
        control_measure_group=None,
        control_measure_map=None,
    ):
        super().__init__(ds, layer_name)
        self.control_group = control_group
        self.control_table = control_table
        self.control_measure_group = control_measure_group
        self.control_measure_map = control_measure_map

    def custom_set_function(self, key, value):
        return key, value

    @property
    def cview(self):
        if not hasattr(self, "control_view"):
            control_view = {}
            for control in self:
                control_view[control.id] = {}

                control_view[control.id]["control"] = control.items
                control_id = control["control_id"]

                # measure groups
                mappings = self.control_measure_map.filter(
                    measure_group_id=control["measure_group_id"]
                )
                control_view[control.id]["measure_group"] = []
                for mapping in mappings:
                    control_view[control.id]["measure_group"].append(
                        {mapping.id: mapping.items}
                    )
                control_view[control.id]["control_table"] = {
                    control_id: self.control_table[control_id].items
                }
                self.control_view = control_view
        return self.control_view


class ControlDelta(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class ControlGroup(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class ControlMeasureGroup(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class ControlMeasureMap(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class ControlMemory(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class ControlPid(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class ControlTable(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class ControlTimed(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class FloodFill(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class Outlet(ThreediVector):

    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class PumpedDrainageArea(ThreediVector):
    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


class InitialWaterlevel(ThreediVector):
    constraints = {}

    def __init__(self, ds, layer_name):
        super().__init__(ds, layer_name)

    def custom_set_function(self, key, value):
        return key, value


MODEL_MAPPING = {
    "v2_simple_infiltration": SimpleInfiltration,
    "v2_global_settings": GlobalSettings,
    "v2_aggregation_settings": AggregationSettings,
    "v2_interflow": Interflow,
    "v2_initial_waterlevel": InitialWaterlevel,
    "v2_groundwater": Groundwater,
    "v2_1d_boundary_conditions": BoundaryConditions1d,
    "v2_2d_boundary_conditions": BoundaryConditions2d,
    "v2_1d_lateral": Lateral1d,
    "v2_2d_lateral": Lateral2d,
    "v2_connection_nodes": ConnectionNodes,
    "v2_cross_section_location": CrossSectionLocation,
    "v2_connected_pnt": ConnectedPnt,
    "v2_calculation_point": CalculationPoint,
    "v2_culvert": Culvert,
    "v2_channel": Channel,
    "v2_pipe": Pipe,
    "v2_weir": Weir,
    "v2_orifice": Orifice,
    "v2_manhole": Manhole,
    "v2_pumpstation": Pumpstation,
    "v2_windshielding": Windshielding,
    "v2_cross_section_definition": CrossSectionDefinition,
    "v2_obstacle": Obstacle,
    "v2_levee": Levee,
    "v2_grid_refinement": GridRefinement,
    "v2_grid_refinement_area": GridRefinementArea,
    "v2_numerical_settings": NumericalSettings,
    "v2_dem_average_area": DemAverageArea,
    "v2_impervious_surface_map": ImperviousSurfaceMap,
    "v2_impervious_surface": ImperviousSurface,
    "v2_surface_parameters": SurfaceParameters,
    "v2_surface_map": SurfaceMap,
    "v2_surface": Surface,
    "v2_control": Control,
    "v2_control_delta": ControlDelta,
    "v2_control_group": ControlGroup,
    "v2_control_measure_group": ControlMeasureGroup,
    "v2_control_measure_map": ControlMeasureMap,
    "v2_control_memory": ControlMemory,
    "v2_control_pid": ControlPid,
    "v2_control_table": ControlTable,
    "v2_control_timed": ControlTimed,
    "v2_floodfill": FloodFill,
    "v2_outlet": Outlet,
    "v2_pumped_drainage_area": PumpedDrainageArea,
    "v2_calculation_point": CalculationPoint,
}



# -*- coding: utf-8 -*-
"""
Created on Fri Sep 16 16:04:45 2022

@author: chris.kerklaan

#TODO:
    Create similar tests as vector.py however add some speed tests.
    
Files used:
    lines.shp
    single.shp
    
"""

# First-party imports
import tempfile
import pathlib

# Third-party imports
from osgeo import ogr

# Local imports
from threedi_edits.gis.vectorbase import VectorBase, load_table

# Globals
# __file__ = "C:/Users/chris.kerklaan/Documents/Github/threedi-raster-edits/threedi_edits/tests/test_gis_vectorbase.py"
DATA = pathlib.Path(__file__).parent / "data" / "gis_vector.gpkg"


def test_load_path():
    """
    Tests if vector is correctly loaded using a path
    A table is also loaded so, it should influence the speed.
    """

    lines = VectorBase(DATA, "lines")
    assert lines.count > 0
    lines.close()


# def test_load_pg():
#     """ tests if vector is correctly loaded using a postgres ds"""
#     pg_test = Vector.from_pg(
#         host="localhost",
#         port=5432,
#         password="postgres",
#         user="postgres",
#         dbname="postgres",
#         layer_name="buurten",
#     )
#     assert pg_test.count > 0
#     pg_test.close()


def test_load_ds():
    """tests if vector is correctly loaded using an ogr datasource"""
    lines_ds = ogr.Open(str(DATA))
    lines = VectorBase.from_ds(lines_ds)
    assert lines.count > 0
    lines.close()


def test_load_table(benchmark):
    lines = VectorBase(DATA, "bgt_panden_fixed")
    layer = lines.layer
    fields = lines.fields

    threshold = 0.05  # or 100 ms or 100.000 us for 3000 features.
    benchmark(load_table, layer, fields.names, fields.types, lines.has_geometry)
    assert benchmark.stats.stats.mean < threshold
    lines.close()


def test_load_scratch():
    """tests if vector can be created from scratch"""
    multi = VectorBase.from_scratch("temp", ogr.wkbMultiPolygon, 28992)
    assert isinstance(multi, VectorBase)


def test_spatial_filtering(benchmark):
    """tests spatial filtering on witin and intersect, with and without layer filter"""

    lines = VectorBase(DATA, "lines")
    lines = VectorBase(lines._copy()[1])

    geometry = VectorBase(DATA, "geometry")[0].geometry

    fids = lines._sfilter(geometry, method="intersects")
    assert len(fids) == 21

    fids = lines._sfilter(geometry, method="intersects")
    assert len(fids) == 21

    fids = lines._sfilter(geometry, method="within")
    assert len(fids) == 14

    fids = lines._sfilter(geometry, method="within")
    assert len(fids) == 14

    threshold = 1 / 1000  # seconds to ms
    benchmark(lines._sfilter, geometry=geometry)
    #     kwargs={"geometry": geometry}, rounds=10
    # )
    assert benchmark.stats.stats.mean < threshold
    lines.close()


def test_table_filter():
    """tests on table filtering
    1. Single filter
    2. double filter
    3. Feature output
    4. fid output
    """

    lines = VectorBase(DATA, "lines")
    lines = VectorBase(lines._copy()[1])
    assert len(VectorBase(lines._filter(HOID="44466"))) == 15
    assert len(VectorBase(lines._filter(HOID="44466", test_field=2))) == 1
    lines.close()


def test_add_feature():
    """tests if a feature is correctly added by using a feature,
    a dictionary and a geometry
    """

    lines = VectorBase(DATA, "lines")
    lines = VectorBase(lines._copy()[1])
    single = VectorBase(DATA, "single")

    lines._add(single[0])
    assert lines.count == 90
    assert lines[89].items["test_field"] == 1

    for value in lines.table.values():
        assert len(value) == 90
    single.close()


def test_delete_feature():
    """tests if feature is properly deleted"""
    lines = VectorBase(DATA, "lines")
    lines = VectorBase(lines._copy()[1])
    lines.delete(0)
    assert lines.count == 88
    assert len(lines.table["fid"]) == 88


def test_copy():
    """tests if features are copied"""
    lines = VectorBase(DATA, "lines")
    copy = VectorBase(lines._copy()[1])
    assert lines.settings == copy.settings
    assert lines.count == copy.count
    assert lines[10].items == copy[10].items
    lines.close()


def test_copy_shell():
    """tests if they have no count"""
    lines = VectorBase(DATA, "lines")
    copy = VectorBase(lines._copy(shell=True)[1])
    assert lines.settings == copy.settings
    assert copy.count == 0
    lines.close()


def test_add_field():
    """tests if field is added and if one can add something"""
    lines = VectorBase(DATA, "lines")
    lines = VectorBase(lines._copy()[1])
    lines.add_field("x", float)
    assert "x" in lines.fields.items

    lines._add(items={"x": 1.0})
    assert lines[89].items["x"] == 1
    lines.close()


def test_delete_field():
    lines = VectorBase(DATA, "lines")
    lines = VectorBase(lines._copy()[1])
    lines.delete_field("test_field")
    assert "test_field" not in lines.fields.items
    assert "test_field" not in lines[80].items
    lines.close()


def test_write_shape():
    """tests if a shape can be written"""

    with tempfile.TemporaryDirectory() as folder:
        lines = VectorBase(DATA, "lines")
        lines = VectorBase(lines._copy()[1])
        lines.write(folder + "/lines_write.shp")
        lines2 = VectorBase(folder + "/lines_write.shp")
        assert lines.count == lines2.count
        assert lines[0].items == lines2[0].items
        assert lines[0].geometry.wkt == lines2[0].geometry.wkt
        lines.close()
        lines2.close()


def test_write_gpkg():
    """tests if a geopackage can be written"""
    with tempfile.TemporaryDirectory() as folder:
        lines = VectorBase(DATA, "lines")
        lines = VectorBase(lines._copy()[1])
        lines.write(folder + "/lines_write.gpkg")
        lines2 = VectorBase(folder + "/lines_write.gpkg")
        assert lines.count == lines2.count
        assert lines[0].items == lines2[1].items
        assert lines[0].geometry.wkt == lines2[1].geometry.wkt
        lines.close()
        lines2.close()



# -*- coding: utf-8 -*-
"""
Created on Tue May 31 10:50:44 2022

@author: chris.kerklaan
"""

# First party imports
import importlib
import logging
from collections import namedtuple

# Local imports
from threedi_edits.globals import INSTALL_REQUIRES

# Globals
logger = logging.getLogger(__name__)
Module = namedtuple(
    "package", "name installed_version installed minimum_required_version"
)


class Dependencies:
    def __init__(self, installs=INSTALL_REQUIRES):
        ogr_module = Module(
            "osgeo.ogr", _check_version("osgeo.ogr"), _check_installed("osgeo.ogr"), 3.4
        )
        self.modules = [ogr_module]
        self._load_requirement_modules(installs)

    def _load_requirement_modules(self, installs):
        for install in installs:
            data = install.split("==")

            if len(data) > 1:
                version = version_to_tuple(data[1])
            else:
                version = None

            module = Module(
                data[0],
                _check_version(data[0]),
                _check_installed(data[0]),
                version,
            )
            self.modules.append(module)
            setattr(self, module.name.replace("-", "_"), module)

        def __repr__(self):
            return self.modules

        def __str__(self):
            return str(self.modules)

    def missing(self):
        for module in self.modules:
            if not module.installed:
                logger.debug(f"Dependency {module} not available!")
                continue

            if module.installed_version and module.minimum_required_version:
                if not module.installed_version >= module.minimum_required_version:
                    logger.debug(f"Dependency {module} version is incorrect!")
                    continue

            # logger.info(f"Dependency {module} correctly installed!")


def _check_installed(module_name):

    try:
        importlib.import_module(module_name)
    except Exception:
        return False
    else:
        return True


def _check_version(module_name):

    try:
        version = importlib.metadata.version(module_name)
        version = version_to_tuple(version)
    except Exception:
        return None
    else:
        return version


def has_ipython() -> bool:
    try:
        get_ipython()  # noqa
    except Exception:
        return False
    return True


def version_to_tuple(version: str):
    return tuple([str(i) for i in version.split(".")])


# Adding a global
DEPENDENCIES = Dependencies()
RUNNING_IPYTHON = has_ipython()



# -*- coding: utf-8 -*-
"""
Created on Tue Apr 30 13:52:55 2019

@author: chris.kerklaan
"""
import importlib


def get_version(version):
    v = importlib.import_module(f"threedi_edits.threedi.constants.v{version}")
    return v



# -*- coding: utf-8 -*-
"""
@author: chris.kerklaan

#TODO functions:
    1. Convert geometry
    2. fix geometry

Currently supported functions
    1. buffer
    2. difference
    3. clip
    4. reproject
    5. to_single

"""

# First-party imports
import pathlib

# Local imports
from threedi_edits.gis.geometry import Geometry
from threedi_edits import Vector

# Globals
DATA = pathlib.Path(__file__).parent / "data" / "gis_geometry.gpkg"
polygon = Vector(DATA, layer_name="geometry")[0].geometry
multiline = Vector(DATA, layer_name="lines").dissolve()[1].geometry


def test_buffer():
    """tests if buffer is correctly performed"""
    geom = Geometry(polygon)
    original_area = geom.area
    buffered = geom.buffer(1)
    assert original_area < buffered.area


def test_difference():
    """tests if difference length is smaller than original length"""
    line = Geometry(multiline)
    differenced = line.difference(polygon)
    assert line.length > differenced.length


def test_clip():
    """tests if clip is correctly performed with correct geom_type"""
    lines = Geometry(multiline)
    clipped = lines.clip(polygon)[0]
    assert clipped.type == 5
    assert clipped.length < lines.length


def test_reproject():
    """tests if geometry is reprojected and correct epsg"""
    geom = Geometry(polygon)
    reprojected = geom.reproject(4326)
    assert reprojected.area < geom.area
    assert reprojected.epsg == 4326


# def test_dissolve():
#     """ tests if ouput of dissolve is 1"""
#     lines = Geometry(multiline)
#     lines.to_single()
#     dissolved = lines.dissolve()


def test_to_single():
    """tests if to single produced multiple geometries"""
    lines = Geometry(multiline)
    single_lines = lines.to_single()
    assert len(single_lines) > 1



# -*- coding: utf-8 -*-
"""
Created on Tue Mar 30 18:53:32 2021

@author: chris.kerklaan
"""
# First-party imports
import pathlib

# Local imports
from threedi_edits.gis.vector import Vector

# Globals
DATA = pathlib.Path(__file__).parent / "data" / "gis_feature.gpkg"


def test_set_feature():
    """tests if feature has been set after field is added"""
    lines = Vector(DATA, layer_name="lines")
    lines = lines.copy()
    lines.add_field("test", float)

    lines[0]["test"] = 0.1

    assert lines[0]["test"] == 0.1
    lines.close()



# -*- coding: utf-8 -*-
"""
Created on Wed Sep 28 13:14:13 2022

@author: chris.kerklaan
"""

# TODO make this
# # First-party imports
# import pathlib

# # Local imports
# from threedi_edits.gis.polygon import Polygon
# from threedi_edits import Vector

# # Globals
# DATA = pathlib.Path(__file__).parent / "data" / "gis_vector.gpkg"
# polygon = Vector(DATA, layer_name="bgt_panden_fixed")[0].geometry.ogr

# def test_centerline():
#     poly = Polygon(polygon)
#     centerline = poly.centerline()
#     empty = Vector.from_scratch("", 3, 28992)
#     empty.add(geometry=centerline)



# -*- coding: utf-8 -*-
"""
Created on Tue Sep  6 14:44:42 2022

@author: chris.kerklaan
"""

# First-party imports
import os
import logging
import operator
import pathlib
import tempfile

# Third party imports
import shapely
from cached_property import cached_property
from osgeo import osr, ogr, _ogr, gdal

# Local imports
from threedi_edits.gis.geometry import (
    TRANSLATION as GEOMETRY_TRANSLATE,
    POLYGON,
    to_ogr as to_ogr_geometry,
    shapely_geometry,
    to_shapely as to_shapely_geometry,
)

from threedi_edits.gis.feature import Feature, to_ogr as to_ogr_feature
from threedi_edits.gis.fields import Fields, GET_FIELD
from threedi_edits.gis.polygon import Polygon
from threedi_edits.gis.spatial_reference import SpatialReference
from threedi_edits.gis.utils import DRIVERS
from threedi_edits.utils.project import Classes, Functions

# structure
classes = Classes(__name__, local_only=True)
functions = Functions(__name__)


# Global DRIVERS
DRIVER_GDAL_MEM = gdal.GetDriverByName("MEM")
DRIVER_OGR_MEM = ogr.GetDriverByName("Memory")
DRIVER_OGR_GPKG = ogr.GetDriverByName("GPKG")
DRIVER_OGR_SHAPEFILE = ogr.GetDriverByName("ESRI Shapefile")
DRIVERS_SQL = ["PostgreSQL", "GPKG"]

# Memory counter
_mem_num = 0

# Logger
logger = logging.getLogger(__name__)


class VectorBase:
    """
    Basewrapper of an ogr layer. This provides the most optimalized
    functions to acces an ogr layer and its data.
    """

    instance = "gis.vector.VectorBase"

    def __init__(
        self,
        path: str = None,
        layer_name: str = None,
        write: int = 1,
        view: bool = False,
        name: str = None,
        epsg: int = 28992,
        datasource: ogr.DataSource = None,
    ):
        path = str(path)  # pathlib input
        self.layer_name = layer_name
        self.view = view
        self.standard_epsg = epsg
        self.memory = "vsimem" in path

        if path is None and datasource is None:
            raise FileNotFoundError("Bath path and datasource are not given.")

        if path is not None and datasource is None:
            self.path = pathlib.Path(path)
            if not self.path.exists() and not self.memory:
                raise FileNotFoundError("Vector not found")
            self.ds = ogr.Open(str(path), write)

        if datasource is not None:
            self.ds = datasource
            self.path = pathlib.Path("")

        if self.path.suffix == ".shp":
            layer_name = self.path.stem

        if layer_name:
            self.layer_name = layer_name
            self.layer = self.ds.GetLayer(layer_name)
            if self.layer is None:
                raise FileNotFoundError(f"{layer_name} does not exist.")
        else:
            try:
                self.layer = self.ds.GetLayer()
                self.layer_name = self.layer.GetName()
            except TypeError:
                raise FileNotFoundError("No layer name, please define")

        if not name:
            name = self.layer_name

        self.name = name
        self.table  # initialize

    @classmethod
    def from_pg(cls, host, port, user, password, dbname, layer_name, write=0, **kwargs):
        """
        returns a class from a postgres database
        takes a {"host":"x", "port":"x", "password":"x",
                 "user":"x", "host":"x"}
        """
        ogr_pg = ("PG:host={} port={} user='{}'" "password='{}' dbname='{}'").format(
            host, port, user, password, dbname
        )

        return cls(
            path=None,
            datasource=ogr.Open(ogr_pg, write),
            layer_name=layer_name,
            **kwargs,
        )

    @classmethod
    def from_ds(cls, datasource: ogr.DataSource = None, **kwargs):
        """returns a class from datasource input"""
        return cls(path=None, datasource=datasource, **kwargs)

    @classmethod
    def from_mem_path(cls, path: str = None, **kwargs):
        """returns a class from a memory path"""
        return cls(path=path, **kwargs)

    @classmethod
    def from_scratch(cls, layer_name: str, geometry_type: int, epsg: int, **kwargs):
        """Creates a layer from scratch
        Params:
            layer_name: name of the layer (str)
            geometry_type: 1 = point, 2 = linestring, 3 = polygon (int)
            epsg: spatial reference code (int)
        """
        return cls(
            path=None,
            datasource=mem_layer(layer_name, geometry_type=geometry_type, epsg=epsg)[0],
        )

    def get_feature(self, fid=None, feature=None):
        """helper for retrieving features"""
        if fid is not None:
            return Feature(
                _ogr.Layer_GetFeature(self.layer, fid), self.layer, self.fields
            )

        if feature:
            return Feature(feature, self.layer, self.fields)

    def __iter__(self):
        self.reset()
        while True:
            feature = _ogr.Layer_GetNextFeature(self.layer)
            if not feature:
                break
            yield self.get_feature(feature=feature)

    def __getitem__(self, i):
        return self.get_feature(i)

    def __len__(self):
        return self.count

    def __setitem__(self, key, value):
        if key == "feature":
            self.set_feature(value)

    def __repr__(self):
        settings = self.settings
        return f"({self.instance}) {self.name} ({len(self)}): {settings}, Type: {self.geometry_name}"

    def __next__(self):
        return self.get_feature(feature=self.layer.GetNextFeature())

    def __in__(self, fid):
        if hasattr(fid, "fid"):
            fid = fid.fid
        return fid in self.fids

    @property
    def count(self):
        return _ogr.Layer_GetFeatureCount(self.layer)

    @property
    def filepath(self):
        return str(self.path)

    @property
    def filename(self):
        return self.path.with_suffix("")

    @property
    def layers(self):
        return [layer.GetName() for layer in self.ds]

    @property
    def extent(self):
        return self.layer.GetExtent()

    @cached_property
    def fields(self):
        return Fields(self.layer)

    @cached_property
    def keys(self):
        return self.fields.keys

    @cached_property
    def field_indices(self):
        return self.fields.indices

    @property
    def fids(self):
        return self.table["fid"]

    @property
    def layer_defn(self):
        return self.layer.GetLayerDefn()

    @property
    def geometry(self):
        return self[0].geometry

    @property
    def geometry_type(self):
        return self.layer.GetGeomType()

    @property
    def geometry_name(self):
        return ogr.GeometryTypeToName(self.geometry_type)

    @property
    def fid_column(self):
        column = self.layer.GetFIDColumn()
        if column == "":
            return "id"
        return self.layer.GetFIDColumn()

    @property
    def driver(self):
        return self.ds.GetDriver().name

    @property
    def spatial_reference(self):
        sr = self.layer.GetSpatialRef()
        if not sr:
            return SpatialReference.from_epsg(self.standard_epsg)
        else:
            return SpatialReference.from_sr(sr)

    @property
    def epsg(self):
        return self.spatial_reference.epsg

    @property
    def extent_geometry(self):
        x1, x2, y1, y2 = self.layer.GetExtent()
        return Polygon.from_wkt(POLYGON.format(x1=x1, x2=x2, y1=y1, y2=y2))

    @property
    def extent_area(self):
        return self.extent_geometry.GetArea()

    @property
    def settings(self):
        return {
            "view": self.view,
            "layer_name": self.layer_name,
            "epsg": self.epsg,
            "name": self.name,
        }

    @property
    def as_gdal_dataset(self):
        """gdal and ogr not always compatible, we have to use openex"""
        mem_ds = mem_path() + ".shp"
        self.write(mem_ds, index=False)
        return gdal.OpenEx(mem_ds)

    @cached_property
    def table(self):
        return load_table(
            self.layer, self.fields.names, self.fields.types, self.has_geometry
        )

    @property
    def has_ogr_index(self):
        return self.layer.TestCapability("FastSpatialFilter")

    @property
    def has_geometry(self):
        return self.geometry_type != 100

    @property
    def geometries(self):
        return self.table["geometry"]

    def reset(self):
        _ogr.Layer_ResetReading(self.layer)

    def first(self):
        return self.get_feature(self.fids[0])

    def add_field(self, name, ogr_type):
        self.fields.add(name, ogr_type)
        self.add_table_field(name)
        self.delete_cached_fields()

    def delete_field(self, *field_names):
        for field_name in field_names:
            self.fields.delete(field_name)
            self.delete_table_field(field_name)

        self.delete_cached_fields()

    def _add(
        self,
        feature=None,
        geometry=None,
        fid=-1,
        fid_latest=False,
        **items,
    ):

        # unpack feature
        if isinstance(feature, Feature):
            geometry = feature.geometry
            items = feature.items

        # pick up items from kwargs
        attr = {}
        if "items" in items:
            attr = items["items"]
            del items["items"]
        attr.update(items)
        items = attr

        # account for none attributes
        if len(items) != 0:
            irrelevant = [k for k in items.keys() if k not in self.keys]
            for key in irrelevant:
                del items[key]

        # convert geometry if needed
        if geometry is not None:
            geometry = to_ogr_geometry(geometry)

        if fid_latest:
            fid = self._get_last_feature_id() + 1

        self.layer, fid = add_feature(
            self.layer,
            self.layer_defn,
            geometry,
            items,
            fid,
        )

        if self.has_geometry:
            items["geometry"] = shapely_geometry(geometry)

        items["fid"] = fid
        self.add_table_feature(items)

        return fid

    def delete(self, feature):
        """deletes by using feature ids
        Input can be a single feature id or of type Feature class
        """

        if type(feature) == int:
            feature = self[feature]

        feature = to_ogr_feature(feature)

        if self.driver == "PostgreSQL":
            self.layer.ResetReading()
            self.ds.StartTransaction()

        self.delete_table_feature(feature)
        self.layer.DeleteFeature(feature.GetFID())

        if self.driver == "PostgreSQL":
            self.ds.CommitTransaction()

    def delete_all(self):
        self.reset()
        for feature in self:
            self.delete(feature)

        self.layer.SyncToDisk()
        self.ds.FlushCache()

        self.layer = None
        self.layer = self.ds.GetLayer(self.layer_name)

    def set_feature(self, feature):
        self.layer.SetFeature(to_ogr_feature(feature))

    def _copy(
        self,
        geometry_type=None,
        shell=False,
        fields=True,
        driver="GPKG",
        spatial_reference=None,
    ) -> (ogr.DataSource, str):
        """In memory ogr object, with features
        params:
            geometry_type: change the geometry type
            shell: Does not add the features
            fields: Set to True is fields have to be included
        """

        if type(geometry_type) == str:
            geometry_type = GEOMETRY_TRANSLATE[geometry_type]

        if not geometry_type:
            geometry_type = self.geometry_type

        ds, layer, path = mem_layer(
            self.layer_name,
            self.layer,
            geometry_type=geometry_type,
            shell=shell,
            fields=fields,
            sr=self.spatial_reference,
            return_path=True,
            driver=driver,
        )
        self.path = path
        return ds, path

    def write(self, path, overwrite=False, driver_name=None):
        """
        Write is done by creating a new file and copying features
        params:
            path: output path of the vector.
            overwrite: Deletes the path is exists.
            driver_name: Short of the ogr driver.
        """
        path = str(path)  # pathlib

        if driver_name is not None:
            driver = ogr.GetDriverByName(driver_name)
        else:
            driver = self.identify_driver(path)
        driver_name = driver.name

        if overwrite:
            if os.path.exists(path):
                driver.DeleteDataSource(path)

        # create output file
        out_ds = driver.CreateDataSource(path)
        out_layer = out_ds.CreateLayer(
            self.layer_name, self.spatial_reference, self.geometry_type
        )

        # add fields
        for i in range(self.layer_defn.GetFieldCount()):
            out_layer.CreateField(self.layer_defn.GetFieldDefn(i))

        # use out layer definition
        layer_defn = out_layer.GetLayerDefn()

        # if uploading to sql based drivers, we have to commit once in a while
        sql = driver_name in DRIVERS_SQL
        if sql:
            out_layer.StartTransaction()

        for i, feature in enumerate(self):
            add_feature(
                out_layer,
                layer_defn,
                to_ogr_geometry(feature.geometry),
                feature.items,
            )
            if sql and (i % 1500 == 0):  # commit once every 1500 features
                out_layer.CommitTransaction()
                out_layer.StartTransaction()

        if sql:
            out_layer.CommitTransaction()

        out_layer = None
        out_ds.Destroy()

    def close(self):
        self.layer = None
        self.ds.Destroy()

    """tables and index"""

    def add_table_feature(self, items):
        for key in self.table:
            if key not in items:
                value = None
            else:
                value = items[key]
            self.__dict__["table"][key].append(value)

    def add_table_field(self, field):
        self.__dict__["table"][field] = [None] * len(self.table["fid"])

    def delete_table_field(self, field):
        del self.__dict__["table"][field]

    def delete_table_feature(self, feature: ogr.Feature):
        """delete a feature in the table"""
        idx = self.table["fid"].index(feature.GetFID())
        for key in self.table:
            del self.__dict__["table"][key][idx]

    def delete_cached_fields(self):
        if "fields" in self.__dict__:
            del self.__dict__["fields"]
        if "keys" in self.__dict__:
            del self.__dict__["keys"]

    """spatial and table filters"""

    def set_spatial_filter(self, geometry):
        self.layer.SetSpatialFilter(geometry)
        self.info(self.layer)

    def _sfilter(self, geometry, method="intersects") -> list:
        """
        Spatial filter of all that is present within the predicates of pygeos.
        https://pygeos.readthedocs.io/en/latest/predicates.html

        Parameters
        ----------
        geometry : shapely.BaseGeometry
            Shapely base geometry.
        method : string
            Intersection method based on the link above.
            The default is "intersects".

        Returns
        -------
        list of ids

        Speed-test with 450.000 features:
            ~32 seconds using this method.
            ~120 with the old method.
            New method is about 4 times faster.

        """
        if not hasattr(shapely.predicates, method):
            raise ValueError(
                f"""
                             Method not in shapely.predicates,
                             please choose from {dir(shapely.predicates)}
                             """
            )
        geometry = to_shapely_geometry(geometry)
        predicate = getattr(shapely.predicates, method)
        data = predicate(self.geometries, geometry).nonzero()
        if len(data[0]) > 0:
            output = operator.itemgetter(*data[0])(self.fids)
            if type(output) not in [list, tuple]:
                return [output]
            return output

        return []

    def _subset(self, *fids) -> str:
        """
        Returns a subset of the vector based on the fids
        fastest way to retrieve fids is via attribute filter and some sql.

        Also when there are more fids that +- 4000 it'll sometimes return an
        error.
        Maybe we should look a settings:
            https://courses.spatialthoughts.com/gdal-tools.html#tips-for-improving-performance
        """

        if len(fids) == 0:
            return self._copy(shell=True)[1]

        if type(fids[0]) == list:
            fids = tuple(fids[0])

        if len(fids) == 1 and type(fids[0]) == int:
            self.layer.SetAttributeFilter(f"FID IN ({fids[0]})")
            return self._copy()[1]

        max_size = 4000
        sets = []
        for i in range(0, len(fids), max_size):
            self.layer.SetAttributeFilter(f"FID IN {fids[i:i + max_size]}")
            sets.append(self._copy()[0])
        output = merge(sets, self.name)

        self.layer.SetAttributeFilter(None)
        return output

    def _filter(self, **filtering) -> str:
        """Filters based  on a dict, returns a Vector"""
        loop_fids = self.table["fid"]
        for field_key, field_value in filtering.items():
            loop_fids = lookup(loop_fids, self.table[field_key], field_value)
        return self._subset(loop_fids)

    def _get_fids(self):
        SQL = f"SELECT fid FROM {self.layer_name}"
        result = self.ds.ExecuteSQL(SQL)
        return [f.GetFID() for f in result]

    def identify_driver(self, path):
        name = DRIVERS[os.path.splitext(path)[1]]
        return ogr.GetDriverByName(name)


def load_table(
    layer: ogr.Layer, field_names: list, field_types: list, has_geometry: bool
) -> dict:
    """
    Retrieves all data from a layer in the quickest way possible.
    We circle around ogr and go directly to the _ogr module.
    This way we are directly accesing the C++ functions.

    Speed-test with 450.000 features:
        ~23 secondds using this method.
        ~163 with the old method.
        New method is about 6 times faster.
        Loading speed increases with more features.

    """

    names = field_names
    if has_geometry:
        extra_fields = ["fid", "geometry"]
    else:
        extra_fields = ["fid"]

    data = {field: [] for field in names + extra_fields}
    indices_types = list(zip(list(range(len(names))), field_types))

    _ogr.Layer_ResetReading(layer)
    feature = _ogr.Layer_GetNextFeature(layer)
    while feature:
        for i, itype in indices_types:
            data[names[i]].append(GET_FIELD[itype](feature, i))

        data["fid"].append(_ogr.Feature_GetFID(feature))
        if has_geometry:
            geometry = _ogr.Feature_GetGeometryRef(feature)

            if geometry is not None:
                data["geometry"].append(shapely_geometry(geometry))
        feature = _ogr.Layer_GetNextFeature(layer)

    return data


def lookup(lookup_table: list, table: list, value: int) -> list:
    """
    Fastest way to lookup a value by using an index in a unordered list
    Looping over a list only once.
    returns a list of matching values.

    A faster way should be with np.searchsorted, however,
    we also have string data.

    """

    values = []

    if value not in table:
        return values

    index = 0

    # last index of value in table
    end = len(table) - operator.indexOf(reversed(table), value) - 1

    # when the first value is the only one
    if end == 0:
        return [lookup_table[0]]

    while index <= end:
        index = table.index(value, index)
        values.append(lookup_table[index])
        index += 1

    return values


def mem_path(extension: str = None) -> str:
    """Generates a new vsimem path with or without an extension"""
    global _mem_num
    location = f"/vsimem/mem{_mem_num}"
    _mem_num += 1
    if extension is not None:
        location = location + f".{extension}"
    else:
        location = location + ".vrt"

    return location


def mem_ds(driver_name: str = "Memory") -> (ogr.DataSource, str):
    """creates a new ogr datasource base on a driver name"""
    driver = ogr.GetDriverByName(driver_name)
    meta = driver.GetMetadata()
    extension = meta.get("DMD_EXTENSION")
    location = mem_path(extension)
    ds = driver.CreateDataSource(location)
    return ds, location


def mem_layer(
    layer_name: str,
    layer: ogr.Layer = None,
    geometry_type: int = None,
    epsg: int = None,
    sr: osr.SpatialReference = None,
    shell: bool = True,
    fields: bool = True,
    options: list = ["SPATIAL_INDEX=YES"],
    return_path: bool = False,
    driver="GPKG",
) -> (ogr.DataSource, ogr.Layer):
    """
    Standard creates a 'shell' of another layer,
    To create a new memory layer,
    or to create a memory copy of a layer
    params:
        epsg/sr: one can choose between epsg/sr
        shell: if True, returns a shell with no  feature
        field: if False, returns a shell without the fields
    returns:
        ds : ogr.Datasource
    """
    ds, path = mem_ds(driver)

    if not shell and layer is not None:
        new_layer = ds.CopyLayer(layer, layer_name)
        if return_path:
            return ds, new_layer, path
        else:
            return ds, new_layer

    if epsg:
        sr = SpatialReference.from_epsg(epsg)
    elif sr == None:
        sr = layer.GetSpatialRef()

    if not geometry_type and layer:
        geometry_type = layer.GetGeomType()
    new_layer = ds.CreateLayer(layer_name, sr, geometry_type, options)

    if fields and layer:
        layer_defn = layer.GetLayerDefn()
        for i in range(layer_defn.GetFieldCount()):
            new_layer.CreateField(layer_defn.GetFieldDefn(i))

    if return_path:
        return ds, new_layer, path
    else:
        return ds, new_layer


def add_feature(
    layer: ogr.Layer,
    feature_defn: ogr.FeatureDefn,
    geometry: ogr.Geometry,
    attributes: dict,
    fid: int = -1,
) -> (ogr.Layer, int):
    """
    Append geometry and attributes as new feature, starting at 0

    for fid=-1 there are three options:
        1. fid is similar to the current count: exisiting fids = [1,4,10], new fid  = 3
        2. fid is 'next number in line': exisiting fids = [1,10], new fid  = 2
        3. fid continues at the latest number: exisiting fids = [1,10], new fid = 11

    """
    feature = ogr.Feature(feature_defn)
    _ogr.Feature_SetFID(feature, fid)

    if geometry is not None:
        _ogr.Feature_SetGeometry(feature, geometry)

    for key, value in attributes.items():
        if value == None:
            _ogr.Feature_SetFieldNull(feature, key)
        else:
            try:
                _ogr.Feature_SetField(feature, key, value)
            except Exception as e:
                raise ValueError("error:", e, "key", key, "value", value)

    try:
        _ogr.Layer_CreateFeature(layer, feature)
    except RuntimeError as e:
        raise RuntimeError("error:", e, "geom:", geometry, "attributes:", attributes)
    finally:
        fid = _ogr.Feature_GetFID(feature)
        _ogr.delete_Feature(feature)

    return layer, fid


def merge(
    datasources: [ogr.DataSource, ogr.DataSource], name="merged"
) -> ogr.DataSource:
    """
    Returns a path for the merged datasources.
    This seems to be the fastest way to merge different vector formats.
    Based on ogrmerge.py and https://gdal.org/drivers/vector/vrt.html
    it uses a vrt combined with a temporary directory to merge the vectors.

    TODO:
        We should lose the temp directory. Maybe we can try this with vsimem.

    """
    with tempfile.TemporaryDirectory() as folder:
        vrt_filename = folder + "/temp.vrt"
        f = gdal.VSIFOpenL(vrt_filename, "wb")
        writer = XMLWriter(f)
        writer.open_element("OGRVRTDataSource")
        gpkg_driver = ogr.GetDriverByName("GPKG")
        writer.open_element("OGRVRTUnionLayer", attrs={"name": name})

        for i, src_ds in enumerate(datasources):
            # first we copy it to the tempfolder
            gpkg_driver.CopyDataSource(src_ds, folder + f"/{i}.gpkg")
            layer = src_ds.GetLayer()
            writer.open_element("OGRVRTLayer", attrs={"name": f"{i}"})
            writer.write_element_value("SrcDataSource", folder + f"/{i}.gpkg")
            writer.write_element_value("SrcLayer", layer.GetName())
            writer.close_element("OGRVRTLayer")

        writer.write_element_value("PreserveSrcFID", "ON")
        writer.close_element("OGRVRTUnionLayer")
        writer.close_element("OGRVRTDataSource")
        gdal.VSIFCloseL(f)

        target = mem_path("gpkg")
        gdal.VectorTranslate(target, vrt_filename, options=["-preserve_fid"])
        gdal.Unlink(vrt_filename)

    return target


# from osgeo utils
def _VSIFPrintfL(f, s):
    gdal.VSIFWriteL(s, 1, len(s), f)


def _Esc(x):
    if gdal.__version__ == "3.5.2":
        return gdal.EscapeString(x, gdal.CPLES_XML).decode("UTF-8")
    else:
        return gdal.EscapeString(x, gdal.CPLES_XML)


class XMLWriter:
    def __init__(self, f):
        self.f = f
        self.inc = 0
        self.elements = []

    def _indent(self):
        return "  " * self.inc

    def open_element(self, name, attrs=None):
        xml_attrs = ""
        if attrs is not None:
            for key in attrs:
                xml_attrs = xml_attrs + ' %s="%s"' % (
                    key,
                    _Esc(attrs[key].encode("utf-8")),
                )
        x = "%s<%s%s>\n" % (self._indent(), name, xml_attrs)
        x = x.encode("utf-8")
        _VSIFPrintfL(self.f, x)
        self.inc = self.inc + 1
        self.elements.append(name)

    def write_element_value(self, name, value, attrs=None):
        xml_attrs = ""
        if attrs is not None:
            for key in attrs:
                xml_attrs = xml_attrs + ' %s="%s"' % (
                    key,
                    _Esc(attrs[key].encode("utf-8")),
                )
        x = "%s<%s%s>%s</%s>\n" % (
            self._indent(),
            name,
            xml_attrs,
            _Esc(value.encode("utf-8")),
            name,
        )
        x = x.encode("utf-8")
        _VSIFPrintfL(self.f, x)

    def close_element(self, closing_name=None):
        self.inc = self.inc - 1
        name = self.elements[-1]
        if closing_name is not None:
            assert name == closing_name
        self.elements = self.elements[0:-1]
        _VSIFPrintfL(self.f, "%s</%s>\n" % (self._indent(), name))



# -*- coding: utf-8 -*-
"""
Created on Fri Apr 26 14:39:17 2019

@author: chris.kerklaan
"""

from . import logging
from . import project

from threedi_edits.utils.project import Files
from threedi_edits.utils.project import Modules

files = Files(__file__)
modules = Modules(__name__)

# pyflakes
logging
project



# -*- coding: utf-8 -*-
"""
Created on Tue Feb 16 16:40:16 2021

@author: chris.kerklaan


ThreediRasterGroup class checks for 3Di properties, alignment etc.

Input:
    1. Raster paths as string
    2. Optional panden

Checks:
    1. Check alignment --> ALready included in rastergroups
    2. Check 3Di properties  --> Done!

TODO:
    - None

Notes:
    1. Memory loading speeds up raster analysis, however is costly in memory
       Not all rasters are loaded into memory, e.g., it is not usefull to load
       a dem into memory since it is not used in conversion.
    2. Maximum memory is three rasters with 500.000.000 pixels each.

"""
# First-party imports
import csv
import shutil
import pathlib
import logging
from pathlib import Path

# Third-party imports
import numpy as np
from osgeo import gdal

# local imports
from threedi_edits.gis.rastergroup import RasterGroup
from threedi_edits.gis.raster import Raster
from threedi_edits.gis.vector import Vector
from threedi_edits.utils.project import Classes, Functions

# structure
classes = Classes(__name__, local_only=True)
functions = Functions(__name__)

# GLOBALS
# Logger
logger = logging.getLogger(__name__)

# CSV path
FILE_PATH = str(pathlib.Path(__file__).parent.absolute()) + "/data/"
CSV_LANDUSE_PATH = FILE_PATH + "Conversietabel_landgebruik_2020.csv"
CSV_SOIL_PATH = FILE_PATH + "Conversietabel_bodem.csv"

# Output paths
OUTPUT_NAMES = {
    "friction": "rasters/friction.tif",
    "dem": "rasters/dem.tif",
    "infiltration": "rasters/infiltration.tif",
    "interception": "rasters/interception.tif",
    "intial_waterlevel": "rasters/initial_waterlevel.tif",
    "infiltration_capacity": "rasters/infiltration_capacity.tif",
    "permeability": "rasters/permeability.tif",
    "max_infiltration_rate": "rasters/max_infiltration_rate.tif",
    "crop_type": "rasters/crop_type.tif",
    "hydraulic_conductivity": "rasters/hydraulic_conductivity.tif",
}
THREEDI_RASTER_NAMES = {
    "dem_file": "dem",
    "interception_file": "interception",
    "frict_coef_file": "friction",
    "infiltration_rate_file": "infiltration",
    "initial_waterlevel_file": "initial_waterlevel",
    "max_infiltration_capacity_file": "infiltration_capacity",
}
THREEDI_RASTER_NAMES_REVERSED = {v: k for k, v in THREEDI_RASTER_NAMES.items()}


CT_LU_KEYS = [
    "Code",
    "Landuse",
    "Friction",
    "Permeability",
    "Interception",
    "Crop_type",
]
CT_LU_TYPES = [int, str, float, float, float, int]

CT_SOIL_KEYS = [
    "Title",
    "Code",
    "Soiltype",
    "Max_infiltration_rate",
    "Hydraulic_conductivity",
]
CT_SOIL_TYPES = [str, int, str, float, float]


class ThreediRasterGroup(RasterGroup):
    def __init__(
        self,
        dem_file: str = None,
        landuse_file: str = None,
        soil_file: str = None,
        interception_file: str = None,
        frict_coef_file: str = None,
        infiltration_rate_file: str = None,
        initial_waterlevel_file: str = None,
        wind_shielding_file: str = None,
        max_infiltration_capacity_file: str = None,
        buildings_file: str = None,
        nodata_value=-9999,
        data_type=gdal.GDT_Float32,
        np_data_type="f4",
        epsg=28992,
        memory=True,
    ):
        """
        ThreediRasterGroup -- info
        ThreediRasterGroup is an object to edit generate,
        check and correct threedi rasters
        General procedure is as follows:
            1. Load rasters by given up the path and create the class group = ThreediRasterGroup(landuse_file='path.tif')
            2. load a conversion csv group.load_landuse_conversion_table()
            3. generate a raster 'group.generate_friction()'
            4. check properties 'group.properties'
            5. check alignment (if you have multiple rasters) 'group.alignment'
            6. correct all mishaps 'group.correct()'
            7. write all rasters  'group.write(folder)'

        Please refer to 'help()' of the specific functions

        params:
            dem_file: str = None,
            landuse_file: str = None,
            soil_file: str = None,
            interception_file: str = None,
            frict_coef_file: str = None,
            infiltration_rate_file: str = None,
            initial_waterlevel_file: str = None,
            max_infiltration_capacity_file: str = None,
            buildings: Vector = None,
            nodata_value=-9999,
            data_type=gdal.GDT_Float32,
            np_data_type="f4",
            epsg=28992
            memory=True

        use self['name'] = Raster to add a raster if it is already an object
        use self.add(path) if you have only a path

        If memory if True, rasters will be loaded into memory.
        If false, they will not. This is slower but needed sometimes to
        preserve memory.
        """

        rasters = []
        self.original_names = {}

        if dem_file:
            logger.info("Loading dem")
            dem = Raster(dem_file, name="dem")
            if memory:
                dem.load_to_memory()
            rasters.append(dem)
            self.original_names["dem"] = Path(dem_file).stem

        if landuse_file:
            logger.info("Loading landuse")
            landuse = Raster(landuse_file, name="landuse")
            rasters.append(landuse)
            self.original_names["landuse"] = Path(landuse_file).stem

        if soil_file:
            logger.info("Loading soil")
            soil = Raster(soil_file, name="soil")
            rasters.append(soil)
            self.original_names["soil"] = Path(soil_file).stem

        if interception_file:
            logger.info("Loading interception")
            interception = Raster(interception_file, name="interception")
            if memory:
                interception.load_to_memory()
            rasters.append(interception)
            self.original_names["interception"] = Path(interception_file).stem

        if frict_coef_file:
            logger.info("Loading friction")
            friction = Raster(frict_coef_file, name="friction")
            if memory:
                friction.load_to_memory()
            rasters.append(friction)
            self.original_names["friction"] = Path(frict_coef_file).stem

        if infiltration_rate_file:
            logger.info("Loading infiltration")
            infiltration = Raster(infiltration_rate_file, name="infiltration")
            if memory:
                infiltration.load_to_memory()
            rasters.append(infiltration)
            self.original_names["infiltration"] = Path(infiltration_rate_file).stem

        if max_infiltration_capacity_file:
            logger.info("Loading infiltration capacity")
            infiltration_capacity = Raster(
                max_infiltration_capacity_file, name="infiltration_capacity"
            )
            if memory:
                infiltration_capacity.load_to_memory()
            rasters.append(infiltration_capacity)
            self.original_names["infiltration_capacity"] = Path(
                max_infiltration_capacity_file
            ).stem

        if initial_waterlevel_file:
            logger.info("Loading initial waterlevel")
            ini_wl = Raster(initial_waterlevel_file, name="initial_waterlevel")
            if memory:
                ini_wl.load_to_memory()
            rasters.append(ini_wl)
            self.original_names["initial_waterlevel"] = Path(
                initial_waterlevel_file
            ).stem

        if wind_shielding_file:
            logger.info("Loading wind shielding")
            wind_shield = Raster(wind_shielding_file, name="wind_shielding")
            if memory:
                wind_shield.load_to_memory()
            rasters.append(wind_shield)
            self.original_names["wind_shielding"] = Path(wind_shielding_file).stem

        RasterGroup.__init__(self, rasters)

        if buildings_file:
            logger.debug("Setting buildings")
            self.buildings = Vector(buildings_file)

        if hasattr(self, "dem"):
            self.epsg = self.dem.epsg
        else:
            self.epsg = epsg
        self.data_type = data_type
        self.no_data_type = np_data_type
        self.nodata_value = nodata_value

        self.retrieve_soil_conversion_table = retrieve_soil_conversion_table
        self.retrieve_landuse_conversion_table = retrieve_landuse_conversion_table

        self.memory = memory

    @property
    def properties(self):
        if not hasattr(self, "_properties"):
            self.check_properties()

        return self._properties

    @property
    def alignment(self):
        if not hasattr(self, "_alignment"):
            self.check_alignment()

        return self._alignment

    @property
    def base_raster(self):
        if hasattr(self, "dem"):
            logger.info("Using dem as base/alignment raster")
            return self.dem
        else:
            logger.info(f"Using the {self.rasters[0].name} as base/alignment raster")
            return self.rasters[0]

    def __str__(self):
        return f"3Di rasters {self.names}"

    def __repr__(self):
        return f"3Di rasters {self.names} \n{self.properties}"

    def add(self, name: str, path: str):
        """adds a raster to the object
        params:
            name: name of raster
            path: path of raster
        """
        if not (name in THREEDI_RASTER_NAMES.values() or name in THREEDI_RASTER_NAMES):
            raise ValueError("name should be object or threedi compatibale")

        if name in THREEDI_RASTER_NAMES:
            name = THREEDI_RASTER_NAMES[name]

        if type(path) == str:
            self[name] = Raster(path)
        else:
            self[name] = path

    def check_table(self, table="soil"):
        """Checks if attribute tables are present, raises an attribute error
        if not present.
        params:
            table: 'soil' or 'landuse'
        """
        logger.info("Checking tables")
        if table == "soil":
            if not hasattr(self, "ct_soil"):
                raise AttributeError(
                    """
                                     Please load soil csv using
                                     'load_soil_conversion_table'"""
                )

        elif table == "landuse":
            if not hasattr(self, "ct_lu"):
                raise AttributeError(
                    """
                                     Please load landuse csv using
                                     'load_landuse_conversion_table'"""
                )

    def fix_table(self):
        """
        Fixes table based on the presence of keys in the current dict
        and fixes the types of the values based on what should be present

        """
        logger.info("Fixing table")

        if hasattr(self, "ct_lu"):
            new_ct_lu = {}
            for key, value in self.ct_lu.items():

                # fix keys
                for fixed_key in CT_LU_KEYS:
                    if fixed_key in key:
                        key = fixed_key

                # fix value types
                key_type = CT_LU_TYPES[CT_LU_KEYS.index(key)]
                new_values = []
                for v in value:
                    if key_type in [float, int]:
                        if isnumber(v):
                            new_values.append(key_type(v))
                    else:
                        new_values.append(key_type(v))

                new_ct_lu[key] = new_values

            self.ct_lu = new_ct_lu

        if hasattr(self, "ct_soil"):
            new_soil_lu = {}
            for key, value in self.ct_soil.items():

                # fix keys
                for fixed_key in CT_LU_KEYS:
                    if fixed_key in key:
                        key = fixed_key

                # fix value types
                key_type = CT_SOIL_TYPES[CT_SOIL_KEYS.index(key)]
                new_values = []
                for v in value:
                    if key_type in [float, int]:
                        if isnumber(v):
                            new_values.append(key_type(v))
                    else:
                        new_values.append(key_type(v))

                new_soil_lu[key] = new_values

            self.ct_soil = new_soil_lu

    def check_properties(
        self, min_allow=-1000, max_allow=1000, max_pixel_allow=1000000000
    ):
        """
        Checks if properties of the rasters are adequate for 3Di
        E.g, epsg is in metre, nodata must be -9999

        params:
            min_allow: minimum allow value of the raster
            max_allow: maximum value of the raster
            max_pixel_allow: total allowed amount of pixels in all rasters
        """
        logger.info("Checking properties")
        self._properties = check_properties(
            self.rasters,
            nodata=self.nodata,
            projection=self.epsg,
            data_type=self.data_type,
            min_allow=min_allow,
            max_allow=max_allow,
            max_pixel_allow=max_pixel_allow,
        )
        return self._properties

    def null_raster(self):
        """Creates a rasters with zeros by using the dem"""
        logger.info("Creating null raster")
        copy = self.base_raster.copy(shell=True, memory=self.memory)
        null_array = np.zeros((int(copy.rows), int(copy.columns)))
        null_array[~self.dem.mask] = np.nan
        copy.array = null_array
        copy.name = "null"
        return copy

    def load_soil_conversion_table(self, csv_soil_path=CSV_SOIL_PATH):
        """loads the soil conversion table, if nothing is entered a standard
        soil table is loaded
        params:
            csv_soil_path
        """
        logger.info("Loading soil conversion table")
        self.ct_soil, self.ct_soil_info = load_csv_file(csv_soil_path, "soil")

    def load_landuse_conversion_table(self, csv_lu_path=CSV_LANDUSE_PATH):
        """loads the soil conversion table, if nothing is entered a standard
        soil table is loaded
        params:
            csv_soil_path
        """
        logger.info("Loading landuse conversion table")
        self.ct_lu, self.ct_lu_info = load_csv_file(csv_lu_path, "landuse")

    def generate_friction(self):
        """Generates a friction raster based on the landuse table"""
        logger.info("Generating friction")
        if hasattr(self, "friction"):
            return

        self.friction = classify(self.landuse, "Friction", self.ct_lu, self.memory)
        self["Friction"] = self.friction

    def generate_permeability(self):
        """Generates a permeability raster based on the landuse table"""
        logger.info("Generating permeability")
        if hasattr(self, "permeability"):
            return
        self.permeability = classify(
            self.landuse, "Permeability", self.ct_lu, self.memory
        )
        self["Permeability"] = self.permeability

    def generate_interception(self):
        """Generates an interception raster based on the landuse table"""
        logger.info("Generating interception")
        if hasattr(self, "interception"):
            return
        self.interception = classify(
            self.landuse, "Interception", self.ct_lu, self.memory
        )
        self["Interception"] = self.interception

    def generate_crop_type(self):
        """Generates a crop type raster based on the landuse table"""
        logger.info("Generating crop type")
        if hasattr(self, "crop_type"):
            return

        self.crop_type = classify(self.landuse, "Crop_type", self.ct_lu, self.memory)
        self["Crop_type"] = self.crop_type

    def generate_max_infiltration(self):
        """Generates a max infiltration raster based on the soil table"""
        logger.info("Generating max infiltration")
        if hasattr(self, "max_infiltration_rate"):
            return

        print("generating max infiltration")
        self.max_infiltration_rate = classify(
            self.soil, "Max_infiltration_rate", self.ct_soil, self.memory
        )
        self["Max_infiltration_rate"] = self.max_infiltration_rate

    def generate_infiltration(self):
        """Generates an infiltration raster based permeability and max infiltration"""
        logger.info("Generating infiltration")
        if hasattr(self, "infiltration"):
            return None

        self.generate_permeability()
        self.generate_max_infiltration()

        output = self.base_raster.copy(shell=True, memory=self.memory)
        for perm_tile, inf_tile in zip(self.permeability, self.max_infiltration_rate):
            perm_array = perm_tile.array
            inf_array = inf_tile.array

            infiltration_array = np.where(
                np.logical_and(
                    perm_array != self.nodata_value,
                    inf_array != self.nodata_value,
                ),
                perm_array * inf_array,
                self.nodata_value,
            ).astype(self.np_data_type)
            output.array = infiltration_array, *perm_tile.location

        self.infiltration = output
        self.infiltration.name = "Infiltration"
        self["Infiltration"] = self.infiltration

    def generate_hydraulic_conductivity(self):
        """Generates a hydraulic conductivity raster based on the soil table"""
        logger.info("Generating hydraulic conductivity")
        if hasattr(self, "hydraulic_conductivity"):
            return
        self.hydraulic_conductivity = classify(
            self.soil, "Hydraulic_conductivity", self.ct_soil, self.memory
        )
        self["Hydraulic_conductivity"] = self.hydraulic_conductivity

    def generate_building_interception(self, value):
        """Generates buildings interception
        params:
            value: value which is placed on top of the buildings
        """
        if hasattr(self, "interception"):
            return

        logger.info("Fixing building vector")
        buildings = self.buildings.fix()

        logger.info("Generating building interception")
        null = self.null_raster()
        value_raster = null.push_vector(buildings, value=value)
        self.interception = value_raster.align(
            self.dem, nodata_align=True, fill_value=0
        )
        self.interception.name = "Interception"
        self["Interception"] = self.interception

    def generate_all(self):
        """generates all rasters except building interception"""

        logger.info("Generating all possible rasters")
        self.generate_friction()
        self.generate_permeability()
        self.generate_interception()
        self.generate_crop_type()
        self.generate_max_infiltration()
        self.generate_infiltration()
        self.generate_hydraulic_conductivity()

    def correct(self, nodata=True, projection=True, data_type=True, alignment=True):
        """Corrects the rasters based on nodata value, epsg, data type and
        Data/Nodata pixels compared to the dem
        params:
            nodata: boolean, if false is skipped
            projection: boolean, if false is skipped
            data_type: boolean, if false is skipped
        """

        if self.count == 0:
            return

        logger.info("Correcting properties of ThreediGroup rasters")
        checks = self.check_properties()

        nodata_value = self.nodata_value
        spatial_reference = self.epsg
        data_type = self.data_type

        for error in checks["errors"]:
            error_type, raster_name, _ = error
            if error_type == "projection" and projection:
                logger.debug(
                    f"replacing projection of {raster_name} to {spatial_reference}"
                )
                raster = getattr(self, raster_name.lower())
                raster.spatial_reference = spatial_reference
                self[raster_name] = raster

            elif error_type == "data_type" and data_type:
                logger.debug(f"replacing data type of {raster_name} to {data_type}")
                raster = getattr(self, raster_name.lower())
                raster.change_data_type(data_type)
                self[raster_name] = raster

        if alignment:
            alignment_check = self.check_alignment()
            for raster in self:
                if raster.name in alignment_check["to_be_aligned"]:
                    logger.debug(f"Aligning {raster.name} to dem")
                    self[raster.name] = raster.align(
                        self.base_raster, idw=False, fill_value=True, nodata_align=True
                    )

        for error in checks["errors"]:
            error_type, raster_name, _ = error
            if error_type == "nodata_value" and nodata:
                logger.debug(f"replacing nodata of {raster_name} to {nodata_value}")
                raster = getattr(self, raster_name.lower())
                raster.replace_nodata(nodata_value, quiet=True)
                self[raster_name] = raster

    def write(self, folder_path, output_names: dict = OUTPUT_NAMES):
        """
           writes raster if present in output names and in object
           creates a folder if it has to...

        params:
           folder_path: path in which the rasters are written
           output_names: e.g. {
           "friction": "rasters/friction.tif",
           "dem":"rasters/dem.tif",
           "infiltration": "rasters/infiltation.tif",
           "interception": "rasters/interception.tif",
           "intial_waterlevel": "rasters/initial_waterlevel.tif"
           }
        """

        # make folder path
        p = pathlib.Path(folder_path)
        p.mkdir(parents=True, exist_ok=True)

        for k, v in output_names.items():
            if v is not None:
                (pathlib.Path(folder_path) / v.split("/")[0]).mkdir(
                    parents=True, exist_ok=True
                )

        for table_name, table_path in output_names.items():
            if table_name in THREEDI_RASTER_NAMES:
                table_name = THREEDI_RASTER_NAMES[table_name]

            if hasattr(self, table_name):
                logger.info(f"Writing {table_name}")
                table = getattr(self, table_name)
                table.write(str(p / table_path))


def isnumber(number):
    """It is quite hard to check if it is a number
    This is currently the best approach
    """
    try:
        float(number)
    except ValueError:
        return False
    else:
        return True


def retrieve_soil_conversion_table(output_path):
    logger.info("retrieving soil conversion table")
    shutil.copyfile(CSV_SOIL_PATH, output_path)


def retrieve_landuse_conversion_table(output_path):
    logger.info("retrieving landuse conversion table")
    shutil.copyfile(CSV_LANDUSE_PATH, output_path)


def check_properties(
    raster_list,
    nodata=-9999,
    projection=28992,
    max_pixel_allow=1000000000,
    data_type=gdal.GDT_Float32,
    unit="metre",
    min_allow=-1000,
    max_allow=1000,
):
    output = {
        "nodata": {},
        "unit": {},
        "projection": {},
        "data_type": {},
        "resolution": {},
        "square_pixels": {},
        "min_max": {},
        "total_pixels": 0,
        "errors": [],
    }

    total_pixels = 0
    for raster in raster_list:

        # nodata value check
        if raster.nodata_value != nodata:
            msg = (
                "nodata_value",
                raster.name,
                f"has a nodata value of {raster.nodata_value}",
            )
            logger.debug(msg)
            output["errors"].append(msg)
        output["nodata"][raster.name] = raster.nodata_value

        raw_nan_count = np.nansum(np.isnan(raster.band.ReadAsArray()))
        if raw_nan_count > 0:
            msg = (
                "nodata_value",
                raster.name,
                " has raw np.nan values",
            )
            logger.debug(msg)
            output["errors"].append(msg)
        output["nodata"][raster.name] = raster.nodata_value

        # unit check
        if raster.spatial_reference.unit != unit:
            msg = ("unit", raster.name, f"has not unit {unit}")
            logger.debug(msg)
            output["errors"].append(msg)
        output["unit"][raster.name] = raster.spatial_reference.unit

        # projection check
        if raster.spatial_reference.epsg != projection:
            msg = ("projection", raster.name, f"has not epsg {projection}")
            logger.debug(msg)
            output["errors"].append(msg)
        output["projection"][raster.name] = raster.spatial_reference.epsg

        # data type check
        if raster.data_type != data_type:
            msg = (
                "data_type",
                raster.name,
                f"is not a {gdal.GetDataTypeName(data_type)}",
            )
            logger.debug(msg)
            output["errors"].append(msg)
        output["data_type"][raster.name] = gdal.GetDataTypeName(data_type)

        # square pixel check
        if abs(raster.resolution["width"]) != abs(raster.resolution["height"]):
            msg = ("width/height", raster.name, "has not a square pixel")
            output["errors"].append(msg)
        output["square_pixels"][raster.name] = raster.resolution

        # extreme value check
        _max, _min = np.nanmax(raster.array), np.nanmin(raster.array)
        if not (min_allow < _max < max_allow and min_allow < _min < max_allow):
            msg = (
                "extreme_values",
                raster.name,
                f"has extreme values < {min_allow}, > {max_allow}",
            )
            logger.debug(msg)
            output["errors"].append(msg)
        output["min_max"][raster.name] = {"min": _min, "max": _max}

        total_pixels += raster.pixels

    # max pixel allowed check
    if total_pixels > max_pixel_allow:
        msg = f"Rasters combined pixels are larger than {max_pixel_allow}"
        logger.debug(msg)
        output["errors"].append(("maximum_allowed_pixels", msg, ""))

    output["total_pixels"] = total_pixels

    if len(output["errors"]) == 0:
        logger.debug("ThreediRasterGroup - Check properties found no problems")
    return output


def load_csv_file(csv_path, csv_type="landuse"):
    csv_data = {}
    csv_info = {}
    if csv_type == "landuse":
        csv_structure = {
            1: "description",
            2: "unit",
            3: "range",
            4: "type",
        }
        meta_list = [1, 2, 3, 4]
    elif csv_type == "soil":
        csv_structure = {
            1: "description",
            2: "source",
            3: "unit",
            4: "range",
            5: "type",
        }
        meta_list = [1, 2, 3, 4, 5]

    with open(csv_path) as csvfile:
        csv_reader = csv.reader(csvfile, delimiter=";")
        for i, line in enumerate(csv_reader):

            # headers
            if i == 0:
                headers = line
                for column_value in line:
                    csv_data[column_value] = []
                    csv_info[column_value] = {}

            # units, descriptions, ranges etc.
            elif i in meta_list:

                for column_index, column_value in enumerate(line):
                    field = csv_structure[i]
                    csv_info[headers[column_index]][field] = column_value

            # csv data
            else:
                for column_index, column_value in enumerate(line):
                    column = headers[column_index]
                    column_type = csv_info[column]["type"]

                    if column_value == "":
                        csv_data[column].append(None)
                    else:
                        if column_type == "Integer":
                            column_value = int(column_value)
                        elif column_type == "Double" or column_type == "Real":
                            column_value = float(column_value)
                        elif column_type == "String":
                            column_value = str(column_value)
                        csv_data[column].append(column_value)

        return csv_data, csv_info


def classify(raster: Raster, table: str, ct: dict, memory=False):
    """input is a table to classify,
    raster is the template raster,
    ct is the conversion table
    returns a dictionary of classified rasters"""

    ct_codes = ct["Code"]
    ct_table = ct[table]
    output = raster.copy(shell=True, memory=memory)
    for tile in raster:
        array = tile.array

        output_array = classify_array(array, ct_table, ct_codes)

        if output_array is None:
            output.array = array, *tile.location
        else:
            output.array = output_array, *tile.location

        tile.close()

    output.name = table
    return output


def classify_array(array, ct_table, ct_codes):
    if np.isnan(array).all():
        return None

    codes = np.unique(array[~np.isnan(array)])
    output_array = np.copy(array)
    for code in codes:
        try:
            output_array[array == code] = ct_table[ct_codes.index(code)]
        except IndexError:
            raise IndexError(f"""{code} not found in table""")
        # array = None # dit levert problemen op

    return output_array



# -*- coding: utf-8 -*-
"""
Created on Thu Feb  4 10:45:53 2021

@author: chris.kerklaan

"""
# Third-party imports
from osgeo import ogr, _ogr

# ogr type mapping
OGR_TYPES = {
    bool: ogr.OFTBinary,
    int: ogr.OFTInteger,
    float: ogr.OFTReal,
    str: ogr.OFTString,
}

OGR_TYPES_INVERT = {
    ogr.OFTInteger: int,  # 0
    ogr.OFSTBoolean: bool,  # 1
    ogr.OFTReal: float,  # 2
    ogr.OFTRealList: float,  # 3
    ogr.OFTString: str,  # 4
    ogr.OFTStringList: str,
    ogr.OFTWideString: str,  # 6
    ogr.OFTBinary: bool,  # 8
    ogr.OFTDate: str,  # 9
    ogr.OFTTime: str,  # 10
    ogr.OFTDateTime: str,  # 11
    ogr.OFTInteger64: int,  # 12
    ogr.OFTInteger64List: int,  # 13
}
GET_FIELD = {
    ogr.OFTInteger: _ogr.Feature_GetFieldAsInteger,  # 0
    ogr.OFSTBoolean: _ogr.Feature_GetFieldAsInteger,  # 1
    ogr.OFTReal: _ogr.Feature_GetFieldAsDouble,  # 2
    ogr.OFTRealList: _ogr.Feature_GetFieldAsDouble,  # 3
    ogr.OFTString: _ogr.Feature_GetFieldAsString,  # 4
    ogr.OFTStringList: _ogr.Feature_GetFieldAsString,
    ogr.OFTWideString: _ogr.Feature_GetFieldAsString,  # 6
    ogr.OFTBinary: _ogr.Feature_GetFieldAsInteger,  # 8
    ogr.OFTDate: _ogr.Feature_GetFieldAsString,  # 9
    ogr.OFTTime: _ogr.Feature_GetFieldAsString,  # 10
    ogr.OFTDateTime: _ogr.Feature_GetFieldAsString,  # 11
    ogr.OFTInteger64: _ogr.Feature_GetFieldAsInteger,  # 12
    ogr.OFTInteger64List: _ogr.Feature_GetFieldAsInteger,  # 13
}


class Fields:
    """ogr fields has to act like a dicionary"""

    def __init__(self, layer):
        self.layer = layer

    def __getitem__(self, field_name):
        return self.type(field_name)

    def __setitem__(self, field_name, field_type):
        self.set_type(field_name, field_type)

    def __repr__(self):
        return str(self.items)

    def __contains__(self, field_name):
        return field_name in self.items

    def __dict__(self):
        return self.items

    @property
    def defn(self):
        if not hasattr(self, "_defn"):
            self._defn = _ogr.Layer_GetLayerDefn(self.layer)
        return self._defn

    @property
    def names(self):
        if not hasattr(self, "_names"):
            self._names = get_field_names(self.defn)
        return self._names

    @property
    def types(self):
        if not hasattr(self, "_types"):
            self._types = get_field_types(self.defn)

        return self._types

    @property
    def nullables(self):
        if not hasattr(self, "_nullables"):
            self._nullables = get_is_nullables(self.defn)
        return self._nullables

    @property
    def type_names(self):
        if not hasattr(self, "_type_names"):
            self._type_names = [ogr.GetFieldTypeName(n) for n in self.types]
        return self._type_names

    @property
    def items(self):
        if not hasattr(self, "_items"):
            self._items = {
                n: [t, tn, ll]
                for n, t, tn, ll in zip(
                    self.names, self.types, self.type_names, self.nullables
                )
            }
        return self._items

    @property
    def keys(self):
        if not hasattr(self, "_keys"):
            self._keys = list(self.items.keys())
        return self._keys

    @property
    def indices(self):
        if not hasattr(self, "_indices"):
            self._indices = {
                name: self.layer.FindFieldIndex(name, 0) for name in self.names
            }
        return self._indices

    def get_field_func(self, field_name):
        """returns a function and an index"""
        return GET_FIELD[self.items[field_name][0]]

    def add(self, field_name, ftype, has_none=True):
        if type(ftype) != int:
            ftype = OGR_TYPES[ftype]

        if field_name not in self.names:
            defn = ogr.FieldDefn(field_name, ftype)
            defn.SetNullable(has_none)
            self.layer.CreateField(defn)

    def delete(self, field_name):
        self.layer.DeleteField(self.layer.FindFieldIndex(field_name, 0))

    def set_name(self, old_field_name, new_field_name):
        i = self.defn.GetFieldIndex(old_field_name)
        field_defn = self.layer.GetFieldDefn(i)
        field_defn.SetName(new_field_name)
        self.layer.AlterFieldDefn(i, field_defn, ogr.ALTER_NAME_FLAG)

    def set_type(self, field_name: str, field_type: type):
        """field type is change by changing the data"""
        ogr_type = OGR_TYPES[field_type]
        data = []

        for feature in self.layer:
            data.append(feature[field_name])

        self.layer.DeleteField(self.layer.FindFieldIndex(field_name, 0))

        defn = ogr.FieldDefn(field_name, ogr_type)
        self.layer.CreateField(defn)

        self.layer.ResetReading()
        for i, feature in enumerate(self.layer):
            value = data[i]
            if value:
                value = field_type(value)

            feature[field_name] = value
            self.layer.SetFeature(feature)


class GeometryFields:
    def __init__(self, feature):
        self.feature = feature

    def __getitem__(self, key):
        return self.items[key]

    @property
    def count(self):
        return _ogr.Feature_GetGeomFieldCount(self.feature)

    @property
    def translate(self):
        return {
            self.ogr_feature.GetGeomFieldDefnRef(i).name: i for i in range(self.count)
        }

    @property
    def names(self):
        return [self.ogr_feature.GetGeomFieldDefnRef(i).name for i in range(self.count)]

    @property
    def items(self):
        geom_items = {}
        for i in range(self.count):
            geometry = self.ogr_feature.GetGeomFieldRef(i)
            geom_items[self.ogr_feature.GetGeomFieldDefnRef(i).name] = geometry
        return geom_items


def get_field_defn(defn):
    return [_ogr.FeatureDefn_GetFieldDefn(defn, i) for i in range(defn.GetFieldCount())]


def get_field_names(defn):
    """returns of list of column names for this field"""
    return [_ogr.FieldDefn_GetName(i) for i in get_field_defn(defn)]


def get_field_types(defn):
    """returns of list of column names for this field"""
    return [_ogr.FieldDefn_GetType(i) for i in get_field_defn(defn)]


def get_is_nullables(defn):
    return [_ogr.FieldDefn_IsNullable(i) for i in get_field_defn(defn)]



# -*- coding: utf-8 -*-
"""
Created on Thu Feb 25 09:56:02 2021

@author: chris.kerklaan

Script for automated testing of the raster object

#TODO functions:
    
Currently supported functions
    1. test_tiling_blocksize
    2. test_tiling_blocksize_optimization
    3. test_merge (used in tiling)
    4. test_tiling (testing changes)
    5. test_push_vector_field
    6. test_push_vector_value
    7. test_push_vector_set_to_nodata
    8. test_empty_copy
    9. align
    10. clip
    11. resample
    12. reproject
    13. idw
    14. polygonize (float)
    15. classify
    16. None memory copies.
    
"""

# First-party imports
import pathlib

# Third-party imports
import numpy as np

# Local imports
from threedi_edits.gis.raster import Raster
from threedi_edits.gis.raster import Vector
from threedi_edits.gis.rastergroup import RasterGroup

# Globals
DATA = pathlib.Path(__file__).parent / "data" / "raster"
VECTOR_DATA = pathlib.Path(__file__).parent / "data" / "gis_raster.gpkg"
VECTOR = Vector(VECTOR_DATA, layer_name="polygon")
RASTER = Raster(DATA / "crop.tif")
RASTER_TEMPLATE = Raster(DATA / "template.tif")


def test_tiling_blocksize():

    RASTER.optimize_blocksize = False
    RASTER.blocksize = (256, 256)

    tiles = [tile for tile in RASTER]

    # total tiles --> only works if the same input is used
    assert len(tiles) == 15

    # First tile
    for tile in RASTER:
        break
    assert tile.size == 256 * 256


def test_tiling_blocksize_optimization():
    """optimization is standard on"""
    RASTER.optimize_blocksize = True
    tiles = [tile for tile in RASTER]

    # total tiles --> only works if the same input is used
    assert len(tiles) == 1


def test_tiling():
    """testing addition in tiling"""
    raster = RASTER.copy()
    raster.optimize_blocksize = False
    raster.blocksize = (256, 256)
    target = raster.copy(shell=True)
    for tile in raster:
        array = tile.array
        tile.array = array + 1
        target.array = array, *tile.location

    assert np.nansum(target.array) == 2803996.0
    target = None


def test_inflated_tiling():
    """tests if inflated tiling works"""
    raster = RASTER.copy()
    for tile, original_extent in raster.tiles_inflated(1, True):
        tile_clipped = tile.clip_bbox(*original_extent)
        break

    for tile_or in raster:
        break

    assert tile_or.shape == tile_clipped.shape
    assert tile.shape[0] - 4 == tile_or.shape[0]  # 4 klopt toch niet?
    assert tile.shape[1] - 4 == tile_or.shape[1]  # 4 klopt toch niet?


def test_push_vector_value():
    """tests push vector with a value"""
    raster = RASTER.copy()
    raster.optimize_blocksize = True
    pushed_raster = raster.push_vector(VECTOR, value=100)
    assert np.nansum(pushed_raster.array) == 2863532.0
    raster = None


def test_push_vector_field():
    """tests push vector when set to field"""
    raster = RASTER.copy()
    pushed_raster = raster.push_vector(VECTOR, field="test")
    assert np.nansum(pushed_raster.array) == 2807372.0
    raster = None


def test_push_vector_nodata():
    """tests push vector when set to nodata"""
    raster = RASTER.copy()
    pushed_raster = raster.push_vector(VECTOR, set_to_nodata=True)
    assert np.nansum(pushed_raster.array) == 2801132.0
    raster = None


def test_copy_shell():
    """tests an emtpy copy, should retrieve a copy of the raster without anything"""
    copy = RASTER.copy(shell=True)

    # array should be 0
    assert np.nansum(copy.array) == 0.0, "Array of empty copy is not 0"

    # check settigns
    assert copy.geotransform == RASTER.geotransform

    # check settigns
    assert copy.nodata_value == RASTER.nodata_value

    # check settigns
    assert copy.spatial_reference.wkt == RASTER.spatial_reference.wkt


def test_too_large_raster():
    """Tests if a too large raster is properly aligned"""

    raster = Raster(DATA / "template_larger.tif")
    group = RasterGroup([raster, RASTER_TEMPLATE])
    checks = group.check_alignment()

    assert len(checks["errors"]) > 0

    aligned = raster.align(RASTER_TEMPLATE, nodata_align=False, fill_value=None)

    group = RasterGroup([aligned, RASTER_TEMPLATE])
    checks = group.check_alignment()

    # Failure can be present in counts, but that's okay due to nodata_align=False
    others = all([False for error in checks["errors"] if error[0] != "counts"])
    assert others


def test_too_small_raster():
    """Tests if a too small raster is properly aligned"""

    raster = Raster(DATA / "template_smaller.tif")
    group = RasterGroup([raster, RASTER_TEMPLATE])
    checks = group.check_alignment()

    assert len(checks["errors"]) > 0

    aligned = raster.align(RASTER_TEMPLATE, nodata_align=False, fill_value=None)

    group = RasterGroup([aligned, RASTER_TEMPLATE])
    checks = group.check_alignment()

    # Failure can be present in counts, but that's okay due to nodata_align=False
    other = all([False for error in checks["errors"] if error[0] != "counts"])
    assert other


def test_raster_w_too_much_nodata():
    """tests if a raster with too much nodata is properly aligned"""

    raster = Raster(DATA / "template_w_nodata.tif")

    group = RasterGroup([raster, RASTER_TEMPLATE])
    checks = group.check_alignment()

    assert len(checks["errors"]) > 0

    aligned = raster.align(RASTER_TEMPLATE, nodata_align=True, fill_value=100)

    group = RasterGroup([aligned, RASTER_TEMPLATE])
    checks = group.check_alignment()

    assert len(checks["errors"]) == 0


def test_raster_w_too_little_nodata():
    """tests if a raster with too little nodata is properly aligned"""
    raster = Raster(DATA / "template_wo_nodata.tif")
    group = RasterGroup([raster, RASTER_TEMPLATE])
    checks = group.check_alignment()

    assert len(checks["errors"]) > 0

    aligned = raster.align(RASTER_TEMPLATE, nodata_align=True, fill_value=100)

    group = RasterGroup([aligned, RASTER_TEMPLATE])
    checks = group.check_alignment()

    assert len(checks["errors"]) == 0


def test_clip():
    """test if a clip is correct"""
    raster = RASTER.copy()
    vector = VECTOR.copy()
    clipped = raster.clip(vector)
    assert np.nansum(clipped.array) == 2856
    raster.close()
    clipped.close()


def test_reproject():
    """tests reproject by geotransform and epsg"""
    raster = RASTER.copy()
    raster.reproject(4326, quiet=False)

    # check epsg
    assert raster.spatial_reference.epsg == 4326, "incorrect epsg"

    # check geotransofrm
    assert raster.geotransform[0] < 1000


def test_resample():
    """tests if resample works"""
    raster = RASTER.copy()
    raster.resample(1, 1, quiet=False)
    assert raster.size == 188340
    assert raster.geotransform[1] == 1.0
    assert raster.geotransform[5] == -1.0


def test_fill_nodata():
    """tests if count of idw result"""
    raster = RASTER.copy()
    edit = raster.copy()
    array = edit.array
    array[array == 8] = raster.nodata_value
    edit.array = array
    idw = edit.fill_nodata(max_iterations=1)
    assert np.nansum(idw.array) == 351146.0


def test_replace_nodata():
    """tests if nodata value of raster is replaced"""
    raster = RASTER.copy()
    is_nodata = ~raster.mask
    raster.replace_nodata(-8888)
    assert raster.nodata_value == -8888

    array = raster.band.ReadAsArray()
    assert (is_nodata == (array == -8888)).all()


def test_polygonize_integer():
    """tests if polygonize works"""
    raster = RASTER.copy()
    raster.change_data_type(int)
    vector = Vector.from_mem_path(raster.polygonize())
    assert vector.table["dn"][0] == 8
    assert type(vector.table["dn"][0]) == int
    assert vector.count == 386


def test_polygonize_float():
    raster = RASTER.copy()
    raster.change_data_type(float)
    vector = Vector.from_mem_path(path=raster.polygonize())
    assert vector.table["dn"][0] == 8
    assert type(vector.table["dn"][0]) == float
    assert vector.count == 386


def test_non_memory_copies():
    raster = RASTER.copy()
    copy = raster.copy(shell=False, memory=False)
    assert copy.ds.GetDriver().LongName == "GeoTIFF"



# -*- coding: utf-8 -*-
"""
Created on Sat May 14 09:28:04 2022

@author: chris.kerklaan

TODO:
    1. geometry grid moeten niet op de directe cellen worden aangesloten,
    maar op de eerste volgende cellen met een waterstand.
"""
# First-party imports
import os
import shutil
import pathlib
import numpy as np

# Third-party imports
import threedi_raster_edits as tre
import xarray as xr

from threedidepth.calculate import calculate_waterdepth
from threedigrid.admin.gridresultadmin import GridH5ResultAdmin


class GridEdit:
    def __init__(self, netcdf_path):
        self.ds = xr.open_dataset(netcdf_path)
        self.edit = self.ds.load()

    def set_wl_timeseries(self, node_id, levels):
        assert len(levels) == len(self.edit["Mesh2D_s1"])

        for timestep in range(len(self.edit["Mesh2D_s1"])):
            self.edit["Mesh2D_s1"][timestep][node_id - 1] = levels[timestep]

    def set_wl_timestep(self, node_id, timestep, level):
        self.edit["Mesh2D_s1"][timestep][node_id - 1] = level

    def write(self, output_path):
        self.edit.to_netcdf(output_path)


# Helper functions
def _grid_geometry(grid: GridH5ResultAdmin, use_ogr=True):
    """returns an ogr Datasource with the cells of a threedigrid"""
    grid.cells.to_shape("/vsimem/cells.shp")
    vector = tre.Vector("/vsimem/cells.shp")
    return vector


def get_neighbor_cells(grid: GridH5ResultAdmin, node_id: int):
    """returns the neighboring cells of a grid cell"""
    cells = _grid_geometry(grid)
    cell = cells[node_id - 1]
    nodes = cells.spatial_filter(cell.geometry)
    return [n["nod_id"] for n in nodes if n["nod_id"] != node_id]


def threedi_nodes_within_vector(vector: tre.Vector, grid: GridH5ResultAdmin):
    """
    returns node ids of 2D and 1D nodes which are present within
    the shape of the Vector.
    params:
        vector: Geopandas geodataframewith polygons
        grid: GridH5ResultAdmin with 1D/2D

    """
    # Get cells
    cells = _grid_geometry(grid)
    within_cells = cells.spatial_filter(vector, method="Within", return_vector=True)
    return within_cells


def edit_nodes_in_grid(
    waterdeel: tre.Vector,
    buildings: tre.Vector,
    threedi_results_folder: str,
    output_folder: str,
    calculation_steps: list,
    dem_height,
):
    """
    writes a new netcdf with an edited version based on the chosen methodology

    This functions edits a threediresult and rewrites 2D waterlevels
    based a the location and a methodology.
    Methodology can be the average of neighboring cells or a 1D replacement.

    params:
        grid:  GridH5ResultAdmin with 1D/2D results.
        vector:gpd.geodataframe.GeoDataFrame
        method:
            - interpolated
            - 1d_replacement

        node_2d_id: 2D node id
        timesteps: waterlevels are returned of this timestep


    """

    # 0. Open variables
    minimum_grid = 10
    search_distance = 10
    cell_search_distance = 2

    netcdf_path = threedi_results_folder + "/results_3di.nc"
    gridadmin_path = threedi_results_folder + "/gridadmin.h5"
    grid = GridH5ResultAdmin(gridadmin_path, netcdf_path)
    edit = GridEdit(netcdf_path)

    # check timestep
    timestep = int(grid.nodes.timestamps[1])
    if timestep not in list(range(300, 340)):
        print("We got a different timestep than 300, we found:", timestep)
        return

    # create folder

    output_folder = pathlib.Path(output_folder)
    output_folder.mkdir(parents=True, exist_ok=True)

    # retrieve the gridcells and clip to the surrounding extent
    water_cell_path = output_folder.parents[0] / "water_cells.shp"
    cell_path = output_folder.parents[0] / "extent_cells.shp"
    if water_cell_path.exists() and cell_path.exists():
        filtered_water_cells = tre.Vector(str(water_cell_path))
        cells = tre.Vector(str(cell_path))
    else:
        grid_cells = _grid_geometry(grid)
        fixed_cells = grid_cells.fix(quiet=False)
        water_cells = fixed_cells.spatial_filter(
            waterdeel.buffer(cell_search_distance), quiet=False, use_ogr=False
        )

        # TODO
        # We have double cells: my_dict = {i:node_ids.count(i) for i in node_ids}

        node_ids = water_cells.table["nod_id"]
        nodes = grid.nodes.filter(id__in=node_ids).timeseries(indexes=calculation_steps)
        node_levels = nodes.s1
        node_ids = nodes.id

        filtered_water_cells = water_cells.copy(shell=True)
        for i in tre.Progress(
            range(0, len(node_ids)), "Filtering water cells on 10 meter levels"
        ):
            node_id = node_ids[i]
            if node_levels[0][i] > dem_height:  # or node_levels[1][i] > dem_height:
                water_cell = water_cells.filter(nod_id=node_id)[0].copy()
                filtered_water_cells.add(water_cell)

        building_cells = fixed_cells.spatial_filter(
            buildings, method="Within", quiet=False, use_ogr=False
        )
        for building_cell in tre.Progress(
            building_cells, "adding within cells of buildings"
        ):
            filtered_water_cells.add(building_cell)

        cells = fixed_cells.spatial_filter(
            filtered_water_cells.buffer(search_distance * 2),
            method="Extent",
            quiet=False,
            use_ogr=False,
        )

        filtered_water_cells.write(str(water_cell_path))
        cells.write(str(cell_path))

    water_grid = tre.Raster(
        filtered_water_cells.rasterize(resolution=minimum_grid, extent=cells.extent)
    )

    water_index = water_grid.array == 1
    levels = {c.id: [] for c in filtered_water_cells}
    for timestep in calculation_steps:
        spatial_timeseries = cells.copy()
        spatial_timeseries.add_field("s1", float)
        timeseries = grid.nodes.timeseries(indexes=[timestep]).s1[0]

        for spatial_timeserie in spatial_timeseries:
            spatial_timeserie["s1"] = float(timeseries[spatial_timeserie["nod_id"]])

        ts_raster = tre.Raster(
            spatial_timeseries.rasterize(resolution=minimum_grid, field="s1")
        )

        # Set all nodata to 0, set all water to np.nan.
        ts_array = ts_raster.array
        ts_array[np.isnan(ts_array)] = 0
        ts_array[water_index] = np.nan
        ts_raster.array = ts_array

        filled = ts_raster.fill_nodata(search_distance, quiet=False)

        ts_raster.write(str(output_folder / f"ts_{timestep}.tif"))
        filled.write(str(output_folder / f"filled_{timestep}.tif"))

        for cell in tre.Progress(
            filtered_water_cells, f"Writing to Netcdf, ts is {timestep}"
        ):
            level = np.nanmean(filled.read(cell.geometry))
            edit.set_wl_timestep(cell["nod_id"], timestep, level)
            levels[cell.id].append(level)

    # write

    interpolated_folder = output_folder / "interpolated"
    interpolated_folder.mkdir(parents=True, exist_ok=True)
    edit.write(str((interpolated_folder / "results_3di.nc")))
    shutil.copy(gridadmin_path, str(output_folder / "interpolated" / "gridadmin.h5"))

    return levels


def calculate_depth(
    threedi_results_folder, dem_path, output_path, calculation_steps: list
):
    """Calculates the waterdepth based on a threedi results
    params:
        folder: A HHNK Folder object
        result_type: "0d1d_result", "1d2d_results" or "batch"

    """

    netcdf_path = threedi_results_folder + "/results_3di.nc"
    gridadmin_path = threedi_results_folder + "/gridadmin.h5"

    for step in calculation_steps:
        path = f"{output_path}/depths_{step}.tif"
        if not os.path.exists(path):
            progress = tre.Progress(total=100, message=f"Creating depths for {step}")

            calculate_waterdepth(
                gridadmin_path,
                netcdf_path,
                dem_path,
                path,
                calculation_steps=[step],
                progress_func=progress.gdal,
                mode="lizard",
            )



# -*- coding: utf-8 -*-
"""
Created on Sat Jul 17 16:02:59 2021

@author: chris

#TODO

#DONE

    # input
    - Create empty sqlite
    - Loading in read mode.
    - Copy a model.
    - Create a grid with a path.
    - Create a grid
    
    # adding 
    - Add a node with items.
    - Add a node with template.
    - add a node with feature
    
    # deleting
    - delete a node
    
    # changing
    - Setting field from a feature.
    - Setting field from a vector.
    - Setting a raster array.
    - Setting a whole new raster.
    - None correcting for memory.
    
    # output
    - Writing from memory.
    
"""
# First-party imports
import pathlib
import tempfile

# Local imports
from threedi_edits.gis.point import Point
from threedi_edits.threedi.edits import ThreediEdits
from threedi_edits.globals import SUPPORTED_THREEDI_VERSIONS
from threedi_edits.threedi.constants import get_version

V = get_version(SUPPORTED_THREEDI_VERSIONS[0])


# Globals
# __file__ = "C:/Users/chris.kerklaan/Documents/Github/threedi-raster-edits/threedi_edits/tests/test_threedi_edits.py"

Templates = V.Templates
DATA = pathlib.Path(__file__).parent / "data" / "threedi_edits"
DEM_PATH = DATA / "dem.tif"

BWN_PATH = DATA / "bwn_test.sqlite"
KLONDIKE_PATH = DATA / "empty_klondike.sqlite"
BWN_SCHEMA_MEMORY = ThreediEdits(BWN_PATH, "memory")


def test_create_empty_model():
    """tests if an empty model can be created"""
    with tempfile.TemporaryDirectory() as folder:

        model = ThreediEdits.from_scratch()
        model.write(folder + "/empty.sqlite")
        model = ThreediEdits(folder + "/empty.sqlite", "memory")

    assert type(model) == ThreediEdits


def test_load_read_mode():
    BWN_SCHEMA = ThreediEdits(BWN_PATH, "read")
    assert BWN_SCHEMA.nodes[15]["code"] == "15"


def test_copy():
    BWN_SCHEMA_COPY = BWN_SCHEMA_MEMORY.copy()
    assert BWN_SCHEMA_COPY.nodes[15]["code"] == "15"


def test_setting_raster():
    """we tests if rasters can be edited and put back in the object"""
    BWN_SCHEMA_MEMORY.rasters  # initialize rasters
    dem = BWN_SCHEMA_MEMORY.dem
    new_dem = dem.copy()
    array = new_dem.array
    array[0, 0] = 2
    new_dem.array = array
    BWN_SCHEMA_MEMORY.dem = new_dem

    assert BWN_SCHEMA_MEMORY.dem.array[0, 0] == 2


def test_create_grid_path():

    BWN_SCHEMA = ThreediEdits(BWN_PATH)
    grid = BWN_SCHEMA.grid()

    assert len(grid["cells"]) == 422


def test_create_grid_memory():
    grid = BWN_SCHEMA_MEMORY.grid()
    assert len(grid["cells"]) == 422


def test_add_delete_node_template():
    templates = Templates()
    node = templates.node
    node["initial_waterlevel"] = 1
    node["the_geom"] = Point.from_point((1, 1))
    new_fid = BWN_SCHEMA_MEMORY.nodes.add(node)

    assert new_fid == 484
    for key in BWN_SCHEMA_MEMORY.nodes.table:
        assert len(BWN_SCHEMA_MEMORY.nodes.table[key]) == 73

    BWN_SCHEMA_MEMORY.nodes.delete(new_fid)

    # deletion in table
    assert BWN_SCHEMA_MEMORY.nodes.fids[-1] == 483
    for key in BWN_SCHEMA_MEMORY.nodes.table:
        assert len(BWN_SCHEMA_MEMORY.nodes.table[key]) == 72


def test_add_items_geometry():
    items = {"initial_waterlevel": 0.1, "storage_are": 0.64, "code": "x"}
    geometry = Point.from_point((1, 1))

    new_fid = BWN_SCHEMA_MEMORY.nodes.add(items=items, geometry=geometry)

    assert new_fid == 484
    for key in BWN_SCHEMA_MEMORY.nodes.table:
        assert len(BWN_SCHEMA_MEMORY.nodes.table[key]) == 73

    BWN_SCHEMA_MEMORY.nodes.delete(new_fid)


def test_add_feature():

    feature = BWN_SCHEMA_MEMORY.nodes[BWN_SCHEMA_MEMORY.nodes.fids[0]]
    new_fid = BWN_SCHEMA_MEMORY.nodes.add(feature)
    assert new_fid == 484
    for key in BWN_SCHEMA_MEMORY.nodes.table:
        assert len(BWN_SCHEMA_MEMORY.nodes.table[key]) == 73

    BWN_SCHEMA_MEMORY.nodes.delete(new_fid)


def test_type_correcting_none():

    # None float
    assert BWN_SCHEMA_MEMORY.global_settings[2]["table_step_size_1d"] is None

    # None int
    assert (
        BWN_SCHEMA_MEMORY.global_settings[1]["simple_infiltration_settings_id"] is None
    )

    # None string
    assert BWN_SCHEMA_MEMORY.global_settings[1]["start_time"] is None


def test_setting_field_from_feature():

    nodes = BWN_SCHEMA_MEMORY.nodes
    node = nodes[15]
    node["code"] = 25

    # first an internal check
    assert BWN_SCHEMA_MEMORY.nodes[15]["code"] == str(25)

    BWN_SCHEMA_MEMORY.nodes[15]["code"] = str(15)


def test_setting_field_from_vector():
    """within this test we retrieve edit a feature and try to put it back"""

    BWN_SCHEMA_MEMORY.nodes[15]["code"] = str(25)

    nodes = BWN_SCHEMA_MEMORY.nodes
    new_nodes = nodes.copy()
    new_nodes[15]["code"] = 26

    # now they should be loose, nothing happens.
    assert BWN_SCHEMA_MEMORY.nodes[15]["code"] == str(25)

    # reset so somthing should happen.
    BWN_SCHEMA_MEMORY.nodes = new_nodes
    assert BWN_SCHEMA_MEMORY.nodes[15]["code"] == str(26)

    BWN_SCHEMA_MEMORY.nodes[15]["code"] = str(15)


def test_setting_raster_array():

    BWN_SCHEMA_MEMORY.rasters

    dem = BWN_SCHEMA_MEMORY.dem

    array = dem.array
    array[0, 0] = 1
    dem.array = array

    assert BWN_SCHEMA_MEMORY.dem.array[0, 0] == 1


def test_write_from_memory():
    """memory causes losses in type checking, so we have to readjust
    and check if everything is written correctly.
    """

    with tempfile.TemporaryDirectory() as tempdir:

        path = str(tempdir) + "/bwn.sqlite"
        BWN_SCHEMA_MEMORY.write(path)

        # check if all fields are correct
        BWN_READ = ThreediEdits(path, mode="read")
        for table in BWN_READ:
            mem_table = BWN_SCHEMA_MEMORY[table.name]
            for feature in table:
                if feature in mem_table:
                    mem_feature = mem_table[feature.id]
                    assert mem_feature.items == feature.items
        table = None
        feature = None
        BWN_READ = None


# def test_node_view():
#     """tests if node view is working"""
#     model = ThreediEdits(BWN_PATH)
#     view = model.node_view
#     assert view[483]["v2_weir"][0] == 10171


# def test_node_delete():
#     model = ThreediEdits(BWN_PATH, "memory")
#     model.delete_node(483)
#     assert 483 not in model.nodes.fids



# -*- coding: utf-8 -*-
"""
Created on Fri May 27 11:22:18 2022

@author: chris.kerklaan
"""
import tempfile
import pathlib
from osgeo import ogr
import logging
import threedi_edits as tre

logger = logging.getLogger(__name__)
from threedi_edits.utils.dependencies import DEPENDENCIES

# Globals
GEOMETRY_FIELDS = ["cells", "lines", "nodes"]

# Third-party imports
if DEPENDENCIES.threedigrid_builder.installed:
    import threedigrid_builder

    def make_grid(model, existing_sqlite_path=None):

        if model.mode == "read":
            sqlite_path = model.path
            dem_path = str(
                pathlib.Path(sqlite_path).parent / f"{model.files['dem_file']}"
            )

        else:
            tempdir = tempfile.TemporaryDirectory()
            temp = pathlib.Path(tempdir.name)
            sqlite_path = temp / "model.sqlite"
            dem_path = temp / f"{model.files['dem_file']}"
            (temp / "rasters").mkdir()

            model.write(sqlite_path, rasters=False, quiet=True)
            model.rasters.dem.write(dem_path)

        grid = threedigrid_builder.make_grid(str(sqlite_path), str(dem_path))
        group = tre.VectorGroup.from_scratch("grid")
        epsg = tre.Raster(dem_path).epsg
        for grid_field in GEOMETRY_FIELDS:
            logger.info(f" Adding {grid_field} to grid.")
            grid_part = grid[grid_field]

            sample_geometry = ogr.CreateGeometryFromWkb(grid_part["geometry"][0])
            sample_geometry_type = sample_geometry.GetGeometryType()

            python_types = {}
            grid_vector = tre.Vector.from_scratch(
                grid_field, sample_geometry_type, epsg
            )
            for field in grid_part:
                if field != "geometry":
                    sample = grid_part[field][0]
                    python_type = getattr(sample, "tolist", lambda: sample)()
                    grid_vector.add_field(field, type(python_type))
                    python_types[field] = type(python_type)

            for i in grid_part["id"]:
                i = i - 1
                items = {}
                for field in grid_part:
                    if field != "geometry":
                        items[field] = python_types[field](grid_part[field][i])

                geometry = ogr.CreateGeometryFromWkb(grid_part["geometry"][i])
                grid_vector.add(fid=i, geometry=geometry, **items)
            group.add(grid_vector, grid_field)

        if model.mode != "read":
            tempdir.cleanup()

        return group



# -*- coding: utf-8 -*-
"""
Created on Tue Feb 23 10:54:38 2021

@author: chris.kerklaan

Functions to test:
#Done:
    # functions
    - rasterize
    - clip 
    - dissolve
    - difference
    - reproject
    - multipart2singlepart
    - centroid
    - simplify

"""
# First-party imports
import sys
import pathlib


# Third-party imports
import numpy as np
from osgeo import ogr

# Local imports
from threedi_edits.gis.vector import Vector
from threedi_edits import Raster

# Globals
# __file__ = "C:/Users/chris.kerklaan/Documents/Github/threedi-raster-edits/threedi_edits/tests/test_gis_vector.py"
DATA = pathlib.Path(__file__).parent / "data" / "gis_vector.gpkg"
RASTER_FOLDER = pathlib.Path(__file__).parent / "data" / "raster"


def test_fix(benchmark):
    """tests if multisurfaces are fixed"""
    bgt = Vector(DATA, layer_name="bgt_panden")
    bgt = bgt.copy()
    fixed = bgt.fix(geometry_type=ogr.wkbMultiPolygon)
    assert fixed.geometry_type == ogr.wkbMultiPolygon
    assert fixed[1].geometry.type == ogr.wkbMultiPolygon

    # benchmark
    # threshold = 0.3
    # benchmark(bgt.fix, geometry_type=ogr.wkbMultiPolygon)
    # assert benchmark.stats.stats.mean < threshold


def test_rasterize_ones():
    """tests rasterization based on counts of array using field"""
    lines = Vector(DATA, layer_name="lines")
    geotransform = (117366.7415, 1.0, 0.0, 459774.6888, 0.0, -1.0)
    rows = 151
    columns = 280
    ds = lines.rasterize(rows, columns, geotransform)
    band = ds.GetRasterBand(1)
    array = band.ReadAsArray()
    array[array == -9999] = np.nan
    assert np.nansum(array) == 1757.0
    lines.close()


def test_rasterize_fields():
    """tests rasterization based on counts of array using field"""

    lines = Vector(DATA, layer_name="lines")
    geotransform = (117366.7415, 1.0, 0.0, 459774.6888, 0.0, -1.0)
    rows = 151
    columns = 280
    ds = lines.rasterize(rows, columns, geotransform, field="test_field")
    band = ds.GetRasterBand(1)
    array = band.ReadAsArray()
    array[array == -9999] = np.nan
    assert np.nansum(array) == 1778.0
    lines.close()


def test_buffer(benchmark):
    """tests if buffer algorithm is performed correctly"""
    lines = Vector(DATA, layer_name="lines")
    buffered = lines.buffer(10, quiet=True)
    assert buffered[1].geometry.area > 0
    lines.close()
    buffered.close()

    # threshold = 0.05
    # benchmark.pedantic(lines.buffer, kwargs={"buffer_size":10}, iterations=10)
    # assert benchmark.stats.stats.mean < threshold


def test_simplify(benchmark):
    """tests if simplify algorithm is performed correctly"""
    shape = Vector(DATA, layer_name="strange_shape")
    simplified = shape.simplify(10)
    assert sys.getsizeof(simplified[1].geometry.wkb) < sys.getsizeof(
        shape[0].geometry.wkb
    )

    # threshold = 0.05
    # benchmark.pedantic(shape.simplify, 10, iterations=10)
    # assert benchmark.stats.stats.mean < threshold


def test_centroid(benchmark):
    """tests if centroid algorithm is performed correctly"""
    lines = Vector(DATA, layer_name="lines")
    centroids = lines.centroid()
    original = lines[0].geometry.Centroid().ExportToWkt()
    new = centroids[1].geometry.centroid.wkt
    assert original == new


def test_multipart_to_singlepart(benchmark):
    """tests if multipart to singlepart on geometry type and counts"""
    multi = Vector(DATA, layer_name="multi")
    singles = multi.to_single()
    assert singles.geometry_type == 3
    assert singles.count == 2
    assert singles[1].geometry.type == 3

    # threshold = 0.05
    # benchmark.pedantic(multi.to_single, iterations=10)
    # assert benchmark.stats.stats.mean < threshold


def test_reproject(benchmark):
    """tests if reprojects has a new bbox, epsg is changed and geometry has changed"""
    lines = Vector(DATA, layer_name="lines")
    new_lines = lines.reproject(4326)
    assert max(new_lines.extent) < 60
    assert new_lines[1].geometry.points[0][0] < 60

    # threshold = 0.05
    # benchmark.pedantic(lines.reproject, 4326, iterations=10)
    # assert benchmark.stats.stats.mean < threshold


def test_clip(benchmark):
    """tests if a clip works by count and feature geometry"""
    lines = Vector(DATA, layer_name="lines")
    mask = Vector(DATA, layer_name="geometry")
    clipped = lines.clip(mask)
    assert clipped.count == 21

    feat_14 = "LINESTRING (117556.454477141 459720.712475407,117576.403681566 459722.13699034)"
    assert clipped[14].geometry.wkt == feat_14

    # threshold = 0.05
    # benchmark.pedantic(lines.clip, mask, iterations=10)
    # assert benchmark.stats.stats.mean < threshold


def test_difference(benchmark):
    """tests if a difference works by count and feature geometry"""
    lines = Vector(DATA, layer_name="lines")
    mask = Vector(DATA, layer_name="geometry")
    differenced = lines.difference(mask)
    assert differenced.count == 75
    feat_51 = "LINESTRING (117575.529616571 459712.808025775,117577.996659787 459713.235393432)"
    assert differenced[52].geometry.wkt == feat_51

    # threshold = 0.05
    # benchmark.pedantic(lines.difference, mask, iterations=10)
    # assert benchmark.stats.stats.mean < threshold


def test_polygon_to_lines():
    """tests polygon to lines"""
    geometry = Vector(DATA, layer_name="geometry")
    lines = geometry.polygon_to_lines()
    assert lines.geometry_type == 2
    assert lines[1].geometry.type == 2


def test_dissolve(benchmark):
    """tests if a difference works by count and wkbsize"""
    lines = Vector(DATA, layer_name="lines")
    dissolved = lines.dissolve()
    assert dissolved.count == 1
    assert dissolved[1].geometry.wkb_size == 3617
    assert dissolved[1].geometry.type == ogr.wkbMultiLineString
    assert dissolved.geometry_type == ogr.wkbMultiLineString

    # threshold = 0.05
    # benchmark.pedantic(lines.dissolve, iterations=10)
    # assert benchmark.stats.stats.mean < threshold


def test_dissolve_field():
    """tests if a difference works by count and wkbsize"""
    lines = Vector(DATA, layer_name="lines")
    dissolved = lines.dissolve(field="HOID")
    assert dissolved.count == 9


def test_integrated_delete_dissolve():
    """tests if integrated works after polygonizing, looping, deleting and getting count
    problem:
        python crashes when this used to be done.

    """
    lines_raster = Raster(RASTER_FOLDER / "lines.tif")
    lines_vector_path = lines_raster.polygonize(quiet=True)
    lines_vector = Vector(lines_vector_path)
    lines_vector = lines_vector.copy()
    count = lines_vector.count

    for i, feature in enumerate(lines_vector):
        if i == 2:
            lines_vector.delete(feature.fid)

    assert lines_vector.count == count - 1

    dissolved = lines_vector.dissolve()
    single_dissolved = dissolved.to_single()
    assert single_dissolved.count == count - 1


def test_voronoi():
    """tests if voronoi can be done"""
    points = Vector(DATA, layer_name="points")
    voronoi = points.voronoi()
    assert voronoi.count == points.count


# def test_spatial_clusters():
#     """  tests if spatial clusters works"""
#     panden = Vector(DATA, layer_name="bgt_panden_fixed")
#     clusters = panden.spatial_clusters()
